# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import Optional

from .step import Step
from ..state import State
from ..state import DesignFormat

from ..logging import info
from ..common import get_script_dir
from ..config import Path, Variable


@Step.factory.register()
class LoadBaseSDC(Step):
    id = "Misc.LoadBaseSDC"
    name = "Load Base SDC"
    long_name = "Load Base Design Constraints File"

    outputs = [DesignFormat.SDC]

    config_vars = [
        Variable(
            "BASE_SDC_FILE",
            Optional[Path],
            "Specifies the base SDC file to source before running Static Timing Analysis.",
            deprecated_names=["SDC_FILE"],
        ),
    ]

    def run(self, **kwargs) -> State:
        new_state = super().run(**kwargs)
        new_state[DesignFormat.SDC] = Path(
            os.path.join(
                get_script_dir(),
                "base.sdc",
            )
        )
        if self.config.get("BASE_SDC_FILE") is not None:
            info(f"Loaded SDC file at '{self.config['BASE_SDC_FILE']}'.")
            new_state[DesignFormat.SDC] = self.config["BASE_SDC_FILE"]
        else:
            info("Loaded default SDC file.")

        return new_state



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum
from typing import Dict


class DesignFormat(Enum):
    """
    A list of design formats that may be kept in the state object.

    The value of the enum is a tuple with the following entries:
        - [0] id
        - [1] file extension
        - [2] human readable name
    """

    NETLIST = (
        "nl",
        "nl.v",
        "Verilog Netlist",
    )
    POWERED_NETLIST = (
        "pnl",
        "pnl.v",
        "Powered Verilog Netlist",
    )

    DEF = (
        "def",
        "def",
        "Design Exchange Format",
    )
    LEF = (
        "lef",
        "lef",
        "Library Exchange Format",
    )
    ODB = (
        "odb",
        "odb",
        "OpenDB Database",
    )

    SDC = (
        "sdc",
        "sdc",
        "Design Constraints",
    )
    SDF = (
        "sdf",
        "sdf",
        "Standard Delay Format",
    )
    SPEF = (
        "spef",
        "spef",
        "Standard Parasitics Extraction Format",
    )
    LIB = (
        "lib",
        "lib",
        "Lib",
    )
    SPICE = (
        "spice",
        "spice",
        "Simulation Program with Integrated Circuit Emphasis",
    )

    GDS = (
        "gds",
        "gds",
        "GDSII Stream",
    )
    MAG_GDS = (
        "mag_gds",
        "magic.gds",
        "GDSII Stream (Magic)",
    )
    KLAYOUT_GDS = (
        "klayout_gds",
        "klayout.gds",
        "GDSII Stream (KLayout)",
    )


DesignFormatByID: Dict[str, DesignFormat] = {
    format.value[0]: format for format in DesignFormat
}



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Common Utilities

A number of common utility functions used throughout the codebase.
"""
import os
import re
import typing
import pathlib
import unicodedata


def mkdirp(path: typing.Union[str, os.PathLike]):
    """
    Attempts to create a directory and all of its parents.

    Does not fail if the directory already exists, however, it does fail
    if it is unable to create any of the components and/or if

    :param path: A filesystem path for the directory
    """
    return pathlib.Path(path).mkdir(parents=True, exist_ok=True)


openlane_root = os.path.dirname(os.path.abspath(__file__))


def get_openlane_root() -> str:
    """
    Returns the root OpenLane folder, i.e., the folder containing the
    ``__init__.py``.
    """
    return openlane_root


def get_script_dir() -> str:
    """
    Gets the OpenLane tool `scripts` directory.
    """
    return os.path.join(
        openlane_root,
        "scripts",
    )


def get_opdks_rev() -> str:
    """
    Gets the Open_PDKs revision.
    """
    return open(os.path.join(openlane_root, "open_pdks_rev"), encoding="utf8").read()


# The following code snippet has been adapted under the following license:
#
# Copyright (c) Django Software Foundation and individual contributors.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.

#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.

#     3. Neither the name of Django nor the names of its contributors may be used
#        to endorse or promote products derived from this software without
#        specific prior written permission.


# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
def slugify(value: str) -> str:
    """
    :param value: Input string
    :returns: The input string converted to lower case, with non-word
        (alphanumeric/underscore) characters removed, and spaces converted
        into hyphens.

        Leading and trailing whitespace is stripped.
    """
    value = (
        unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    )
    value = re.sub(r"[^\w\s\-\.]", "", value).strip().lower()
    return re.sub(r"[-\s\.]+", "-", value)


final = typing.final


def internal(method):
    """A decorator to indicate internal methods.

    It dynamically adds a statement to the effect in the docstring, but has no
    other runtime effects.

    :param f: Method to mark as internal
    """
    if method.__doc__ is None:
        method.__doc__ = ""
    method.__doc__ = (
        "**This method is considered internal and should only be called by this class or its subclass hierarchy.**\n"
        + method.__doc__
    )
    return method



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import os
import glob
import datetime
from abc import abstractmethod, ABC
from concurrent.futures import Future, ThreadPoolExecutor
from typing import (
    List,
    Sequence,
    Tuple,
    Type,
    ClassVar,
    Optional,
    Dict,
    Callable,
    Union,
)

from rich.progress import (
    Progress,
    TextColumn,
    BarColumn,
    MofNCompleteColumn,
    TimeElapsedColumn,
    TaskID,
)

from ..config import (
    Config,
    ConfigBuilder,
    Variable,
    universal_flow_config_variables,
)
from ..state import State
from ..steps import Step
from ..utils import Toolbox
from ..logging import console, info
from ..common import mkdirp, internal, final, slugify


class FlowException(RuntimeError):
    pass


class FlowError(RuntimeError):
    pass


class FlowImplementationError(RuntimeError):
    pass


class Flow(ABC):
    """
    An abstract base class for a flow.

    Flows encapsulates a subroutine that runs multiple steps: either synchronously,
    asynchronously, serially or in any manner.

    The Flow ABC offers a number of convenience functions, including handling the
    progress bar at the bottom of the terminal, which shows what stage the flow
    is currently in and the remaining stages.

    :param config: A resolved OpenLane configuration object to be used as input
        for this instance of a flow.
    :param design_dir: A design directory
    :param name: An optional string name for the Flow itself, and not a run of it.

        If not set, the Python class name will be used instead.
    :param Steps:
        A list of :class:`Step` **types** used by the Flow (not Step objects.)

        Subclasses of :class:`Step` are expected to override the default value
        as a class member.
    """

    name: str
    Steps: List[Type[Step]] = []  # Override

    def __init__(
        self,
        config: Config,
        design_dir: str,
        name: Optional[str] = None,
        Steps: Optional[List[Type[Step]]] = None,
    ):
        if Steps is not None:
            self.Steps = Steps

        self.name = name or self.__class__.__name__
        self.config: Config = config
        self.design_dir = design_dir

        self.tpe: ThreadPoolExecutor = ThreadPoolExecutor()

        self.ordinal: int = 1
        self.completed: int = 0
        self.max_stage: int = 0
        self.task_id: Optional[TaskID] = None
        self.progress: Optional[Progress] = None
        self.run_dir: Optional[str] = None
        self.tmp_dir: Optional[str] = None
        self.toolbox: Optional[Toolbox] = None

    @classmethod
    def get_config_variables(Self) -> List[Variable]:
        flow_variables_by_name: Dict[str, Tuple[Variable, str]] = {
            variable.name: (variable, "Universal")
            for variable in universal_flow_config_variables
        }
        for step_cls in Self.Steps:
            for variable in step_cls.config_vars:
                if flow_variables_by_name.get(variable.name) is not None:
                    existing_variable, existing_step = flow_variables_by_name[
                        variable.name
                    ]
                    if variable != existing_variable:
                        raise FlowException(
                            f"Misconfigured flow: Unrelated variables in {existing_step} and {step_cls.__name__} share a name: {variable.name}"
                        )
                flow_variables_by_name[variable.name] = (variable, step_cls.__name__)

        return [variable for variable, _ in flow_variables_by_name.values()]

    @classmethod
    def init_with_config(
        Self,
        config_in: Union[str, os.PathLike, Dict],
        pdk: Optional[str] = None,
        pdk_root: Optional[str] = None,
        scl: Optional[str] = None,
        design_dir: Optional[str] = None,
        config_override_strings: Optional[Sequence[str]] = None,
        **kwargs,
    ):
        """
        Create a new instance of the :class:`Flow` object with a given
        configuration file or object.

        :param config_in: See :meth:`ConfigBuilder.load`
        :param config_override_strings: See :meth:`ConfigBuilder.load`
        :param pdk: See :meth:`ConfigBuilder.load`
        :param pdk_root: See :meth:`ConfigBuilder.load`
        :param scl: See :meth:`ConfigBuilder.load`
        :param design_dir: See :meth:`ConfigBuilder.load`
        """

        flow_config_vars = Self.get_config_variables()

        config_resolved, design_dir_resolved = ConfigBuilder.load(
            config_in=config_in,
            flow_config_vars=flow_config_vars,
            config_override_strings=config_override_strings,
            pdk=pdk,
            pdk_root=pdk_root,
            scl=scl,
            design_dir=design_dir,
        )

        return Self(
            config=config_resolved,
            design_dir=design_dir_resolved,
            **kwargs,
        )

    @final
    def start(
        self,
        with_initial_state: Optional[State] = None,
        tag: Optional[str] = None,
        **kwargs,
    ) -> Tuple[State, List[Step]]:
        """
        The entry point for a flow.

        :param with_initial_state: An optional initial state object to use.
            If not provided:
            * If resuming a previous run, the latest `state_out.json` (by filesystem modification date)
            * If not, an empty state object is created.
        :param tag: A name for this invocation of the flow. If not provided,
            one based on a date string will be created.

        :returns: `(success, state_list)`
        """
        if tag is None:
            tag = datetime.datetime.now().astimezone().strftime("RUN_%Y-%m-%d_%H-%M-%S")

        self.run_dir = os.path.join(self.design_dir, "runs", tag)
        self.tmp_dir = os.path.join(self.run_dir, "tmp")
        self.toolbox = Toolbox(self.tmp_dir)

        initial_state = with_initial_state or State()

        try:
            entries = os.listdir(self.run_dir)
            if len(entries) == 0:
                raise FileNotFoundError(self.run_dir)  # Treat as non-existent directory
            info(f"Using existing run at '{tag}' with the '{self.name}' flow.")

            # Extract maximum step ordinal
            for entry in entries:
                components = entry.split("-")
                if len(components) < 2:
                    continue
                try:
                    current_ordinal = int(components[0])
                except ValueError:
                    continue
                self.ordinal = max(self.ordinal, current_ordinal + 1)

            # Extract Maximum State
            if with_initial_state is None:
                latest_time: float = 0
                latest_json: Optional[str] = None
                state_out_jsons = glob.glob(
                    os.path.join(self.run_dir, "**", "state_out.json"), recursive=True
                )
                for state_out_json in state_out_jsons:
                    time = os.path.getmtime(state_out_json)
                    if time > latest_time:
                        latest_time = time
                        latest_json = state_out_json

                if latest_json is not None:
                    initial_state = State.loads(
                        open(latest_json, encoding="utf8").read()
                    )

        except NotADirectoryError:
            raise FlowException(
                f"Run directory for '{tag}' already exists as a file and not a directory."
            )
        except FileNotFoundError:
            info(f"Starting a new run of the '{self.name}' flow with the tag '{tag}'.")
            mkdirp(self.run_dir)

        config_res_path = os.path.join(self.run_dir, "resolved.json")
        with open(config_res_path, "w") as f:
            f.write(self.config.dumps())

        self.progress = Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            MofNCompleteColumn(),
            TimeElapsedColumn(),
            console=console,
        )
        self.progress.start()
        self.task_id = self.progress.add_task(
            f"{self.name}",
        )
        result = self.run(
            initial_state=initial_state,
            **kwargs,
        )
        self.progress.stop()

        # Reset stateful objects
        self.progress = None
        self.task_id = None
        self.tmp_dir = None
        self.toolbox = None
        self.ordinal = 1
        self.completed = 0
        self.max_stage = 0

        return result

    @internal
    @abstractmethod
    def run(
        self,
        initial_state: State,
        **kwargs,
    ) -> Tuple[State, List[Step]]:
        """
        The core of the Flow. Subclasses of flow are expected to override this
        method.

        :param initial_state: An initial state object to use.
        :returns: A tuple of states and instantiated step objects for inspection.
        """
        pass

    @internal
    def start_step_async(
        self,
        step: Step,
        toolbox: Optional[Toolbox] = None,
        *args,
        **kwargs,
    ) -> Future[State]:
        """
        A helper function that may run a step asynchronously.

        It returns a `Future` encapsulating a State object, which can then be
        used as an input to the next step or inspected to await it.

        See the Step initializer for more info.

        :param step: The step object to run
        :param toolbox: An optional override for the flow's :class:`Toolbox`.
            If you don't know what this means, just leave this as ``None``.
        :param args: Arguments to `step.start`
        :param kwargs: Keyword arguments to `step.start`
        """

        kwargs["toolbox"] = toolbox or self.toolbox

        return self.tpe.submit(step.start, *args, **kwargs)

    @internal
    def set_max_stage_count(self, count: int):
        """
        A helper function, used to set the total number of stages a flow is
        expected to go through. Used to set the progress bar.

        :param count: The total number of stages.
        """
        if self.progress is None or self.task_id is None:
            return
        self.max_stage = count
        self.progress.update(self.task_id, total=count)

    @internal
    def start_stage(self, name: str):
        """
        Starts a new stage, updating the progress bar appropriately.

        :param name: The name of the stage.
        """
        if self.progress is None or self.task_id is None:
            return
        self.progress.update(
            self.task_id,
            description=f"{self.name} - Stage {self.ordinal} - {name}",
        )

    @internal
    def end_stage(self, no_increment_ordinal: bool = False):
        """
        Ends the current stage, updating the progress bar appropriately.
        """
        self.completed += 1
        if not no_increment_ordinal:
            self.ordinal += 1
        assert self.progress is not None and self.task_id is not None
        self.progress.update(self.task_id, completed=float(self.completed))

    @internal
    def current_stage_prefix(self) -> str:
        """
        Returns a prefix for a step ID with its stage number so it can be used
        to create a step directory.
        """
        max_stage_digits = len(str(self.max_stage))
        return f"%0{max_stage_digits}d-" % self.ordinal

    @internal
    def dir_for_step(self, step: Step):
        """
        Returns a directory within the run directory for a specific step,
        prefixed with the current progress bar stage number.
        """
        if self.run_dir is None:
            raise FlowImplementationError(
                "Attempted to call dir_for_step before starting flow."
            )
        return os.path.join(
            self.run_dir,
            f"{self.current_stage_prefix()}{slugify(step.id)}",
        )

    class FlowFactory(object):
        """
        A factory singleton for Flows, allowing Flow types to be registered and then
        retrieved by name.

        See https://en.wikipedia.org/wiki/Factory_(object-oriented_programming) for
        a primer.
        """

        _registry: ClassVar[Dict[str, Type[Flow]]] = {}

        @classmethod
        def register(
            Self, registered_name: Optional[str] = None
        ) -> Callable[[Type[Flow]], Type[Flow]]:
            """
            A decorator that adds a flow type to the registry.

            :param registered_name: An optional registered name for the flow.

                If not specified, the flow will be referred to by its Python
                class name.
            """

            def decorator(cls: Type[Flow]) -> Type[Flow]:
                name = cls.__name__
                if registered_name is not None:
                    name = registered_name
                Self._registry[name] = cls
                return cls

            return decorator

        @classmethod
        def get(Self, name: str) -> Optional[Type[Flow]]:
            """
            Retrieves a Flow type from the registry using a lookup string.

            :param name: The registered name of the Flow. Case-sensitive.
            """
            return Self._registry.get(name)

        @classmethod
        def list(Self) -> List[str]:
            """
            :returns: A list of strings representing all registered flows.
            """
            return list(Self._registry.keys())

    factory = FlowFactory
    get = FlowFactory.get



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import re
import sys
import json
from abc import abstractmethod
from typing import List, Optional


from .step import Step
from ..state import State
from ..state import DesignFormat
from .common_variables import io_layer_variables

from ..config import Path, Variable
from ..common import get_openlane_root, get_script_dir

inf_rx = re.compile(r"\b(-?)inf\b")


class OdbpyStep(Step):
    inputs = [DesignFormat.ODB]
    outputs = [DesignFormat.ODB, DesignFormat.DEF]

    def run(
        self,
        **kwargs,
    ) -> State:
        """
        TODO
        """

        state_out = super().run(**kwargs)
        kwargs, env = self.extract_env(kwargs)

        out_paths = {}

        command = self.get_command()
        for output in [DesignFormat.ODB, DesignFormat.DEF]:
            id, ext, _ = output.value
            filename = f"{self.config['DESIGN_NAME']}.{ext}"
            file_path = os.path.join(self.step_dir, filename)
            command.append(f"--output-{id}")
            command.append(file_path)
            out_paths[output] = Path(file_path)

        command += [
            str(state_out[DesignFormat.ODB]),
        ]

        env["OPENLANE_ROOT"] = get_openlane_root()
        env["ODB_PYTHONPATH"] = ":".join(sys.path)

        self.run_subprocess(
            command,
            env=env,
            **kwargs,
        )

        metrics_path = os.path.join(self.step_dir, "or_metrics_out.json")
        if os.path.exists(metrics_path):
            metrics_str = open(metrics_path).read()
            metrics_str = inf_rx.sub(lambda m: f"{m[1] or ''}\"Infinity\"", metrics_str)
            new_metrics = json.loads(metrics_str)
            state_out.metrics.update(new_metrics)

        for output in [DesignFormat.ODB, DesignFormat.DEF]:
            state_out[output] = out_paths[output]

        return state_out

    def get_command(self) -> List[str]:
        metrics_path = os.path.join(self.step_dir, "or_metrics_out.json")
        lefs = ["--input-lef", self.config["TECH_LEF"]]
        for lef in self.config["CELL_LEFS"]:
            lefs.append("--input-lef")
            lefs.append(lef)
        if extra_lefs := self.config["EXTRA_LEFS"]:
            for lef in extra_lefs:
                lefs.append("--input-lef")
                lefs.append(lef)
        return [
            "openroad",
            "-exit",
            "-metrics",
            metrics_path,
            "-python",
            self.get_script_path(),
        ] + lefs

    @abstractmethod
    def get_script_path(self):
        pass


@Step.factory.register()
class ApplyDEFTemplate(OdbpyStep):
    id = "Odb.ApplyDEFTemplate"
    name = "Apply DEF Template"

    flow_control_variable = "FP_DEF_TEMPLATE"
    flow_control_msg = "No DEF template provided, skipping…"

    config_vars = [
        Variable(
            "FP_DEF_TEMPLATE",
            Optional[Path],
            "Points to the DEF file to be used as a template when running `apply_def_template`. This will be used to exctract pin names, locations, shapes -excluding power and ground pins- as well as the die area and replicate all this information in the `CURRENT_DEF`.",
        ),
    ]

    def get_script_path(self):
        return os.path.join(
            get_script_dir(),
            "odbpy",
            "apply_def_template.py",
        )

    def get_command(self) -> List[str]:
        return super().get_command() + [
            "--def-template",
            self.config["FP_DEF_TEMPLATE"],
        ]


@Step.factory.register()
class ManualMacroPlacement(OdbpyStep):
    id = "Odb.ManualMacro"
    name = "Manual Macro Placement"

    flow_control_variable = "MACRO_PLACEMENT_CFG"
    flow_control_msg = "No macros configured, skipping…"

    config_vars = [
        Variable(
            "MACRO_PLACEMENT_CFG",
            Optional[Path],
            "Specifies the path a file specifying how OpenLane should place certain macros.",
        ),
    ]

    def get_script_path(self):
        return os.path.join(get_script_dir(), "odbpy", "manual_macro_place.py")

    def get_command(self) -> List[str]:
        return super().get_command() + [
            "--config",
            self.config["MACRO_PLACEMENT_CFG"],
            "--fixed",
        ]


@Step.factory.register()
class ReportWireLength(OdbpyStep):
    id = "Odb.ReportWireLength"
    name = "Report Wire Length"

    def get_script_path(self):
        return os.path.join(get_script_dir(), "odbpy", "wire_lengths.py")

    def get_command(self) -> List[str]:
        return super().get_command() + [
            "--human-readable",
            "--report-out",
            os.path.join(self.step_dir, "wire_lengths.csv"),
        ]


@Step.factory.register()
class CustomIOPlacement(OdbpyStep):
    id = "Odb.CustomIOPlacement"
    name = "Custom I/O Placement"
    long_name = "Custom I/O Pin Placement Script"

    flow_control_variable = "FP_PIN_ORDER_CFG"
    flow_control_msg = "No custom floorplan file configured, skipping…"

    config_vars = io_layer_variables + [
        Variable(
            "FP_PIN_ORDER_CFG",
            Optional[Path],
            "Points to the pin order configuration file to set the pins in specific directions (S, W, E, N). If not set, then the IO pins will be placed using OpenROAD's basic pin placer.",
        ),
        Variable(
            "QUIT_ON_UNMATCHED_IO",
            bool,
            "Exit on unmatched pins in a provided `FP_PIN_ORDER_CFG` file.",
            default=True,
            deprecated_names=["FP_IO_UNMATCHED_ERROR"],
        ),
    ]

    def get_script_path(self):
        return os.path.join(get_script_dir(), "odbpy", "io_place.py")

    def get_command(self) -> List[str]:
        length = max(
            self.config["FP_IO_VLENGTH"],
            self.config["FP_IO_HLENGTH"],
        )

        return super().get_command() + [
            "--config",
            self.config["FP_PIN_ORDER_CFG"],
            "--hor-layer",
            self.config["FP_IO_HLAYER"],
            "--ver-layer",
            self.config["FP_IO_VLAYER"],
            "--hor-width-mult",
            str(self.config["FP_IO_VTHICKNESS_MULT"]),
            "--ver-width-mult",
            str(self.config["FP_IO_HTHICKNESS_MULT"]),
            "--hor-extension",
            str(self.config["FP_IO_HEXTEND"]),
            "--ver-extension",
            str(self.config["FP_IO_VEXTEND"]),
            "--length",
            str(length),
            (
                "--unmatched-error"
                if self.config["QUIT_ON_UNMATCHED_IO"]
                else "--ignore-unmatched"
            ),
        ]



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import os
import re
import glob
import shutil
import textwrap
import subprocess
from enum import Enum
from decimal import Decimal
from collections import deque
from typing import List, Dict, Sequence, Union
from os.path import join, dirname, isdir, relpath

from .step import Step
from ..state import State
from ..config import Path
from ..logging import info, warn
from ..common import mkdirp, get_script_dir, get_openlane_root


def create_reproducible(
    step_dir: str,
    cmd: Sequence[Union[str, os.PathLike]],
    env_in: Dict[str, str],
    tcl_script: str,
    verbose: bool = False,
) -> str:
    step_dir = os.path.abspath(step_dir)
    run_path = os.path.dirname(step_dir)
    run_path_rel = os.path.relpath(run_path)
    pdk_root = env_in["PDK_ROOT"]

    info(
        textwrap.dedent(
            """\
            OpenLane TCLStep Issue Packager

            EFABLESS CORPORATION AND ALL AUTHORS OF THE OPENLANE PROJECT SHALL NOT BE HELD
            LIABLE FOR ANY LEAKS THAT MAY OCCUR TO ANY PROPRIETARY DATA AS A RESULT OF USING
            THIS SCRIPT. THIS SCRIPT IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OR
            CONDITIONS OF ANY KIND.

            BY USING THIS SCRIPT, YOU ACKNOWLEDGE THAT YOU FULLY UNDERSTAND THIS DISCLAIMER
            AND ALL IT ENTAILS.
            """
        )
    )

    # Phase 1: Set up destination folder
    destination_folder = os.path.join(step_dir, "tcl_reproducible")
    info(f"Setting up '{destination_folder}'…")

    try:
        shutil.rmtree(destination_folder)
    except FileNotFoundError:
        pass

    mkdirp(destination_folder)

    # Phase 2: Process TCL Scripts To Find Full List Of Files
    tcl_script_abspath = os.path.abspath(tcl_script)

    script_counter = 0

    def get_script_key():
        nonlocal script_counter
        value = f"PACKAGED_SCRIPT_{script_counter}"
        script_counter += 1
        return value

    tcls_to_process = deque([(tcl_script, get_script_key())])

    new_cmd = []
    for element in cmd:
        element = str(element)
        element_basename = os.path.basename(element)
        element_abspath = os.path.abspath(element)
        if os.path.exists(element):
            if element_abspath == tcl_script_abspath:
                new_cmd.append("$PACKAGED_SCRIPT_0")
            elif element_abspath.endswith(".tcl") or element_abspath.endswith(
                ".magicrc"
            ):
                env_key = get_script_key()
                tcls_to_process.append((element_abspath, env_key))
                new_cmd.append(f"${env_key}")
            else:
                new_cmd.append(element)
        elif element.startswith(run_path_rel) or element_abspath.startswith(run_path):
            new_cmd.append(element_basename)
        else:
            new_cmd.append(element)

    def shift(deque):
        try:
            return deque.popleft()
        except Exception:
            return None

    env_keys_used = set(["PDK_ROOT"])
    tcls = set()
    env = env_in.copy()
    current = shift(tcls_to_process)
    loop_guard = 1024
    while current is not None:
        env_key = None
        if isinstance(current, tuple):
            current, env_key = current

        loop_guard -= 1
        if loop_guard == 0:
            raise Exception("An infinite loop has been detected. Please file an issue.")

        if env_key is None:
            env_key = get_script_key()

        env_keys_used.add(env_key)
        env[env_key] = current

        warnings = []

        try:
            script = open(current).read()
            if verbose:
                info(f"Processing {current}…")

            for key, value in env.items():
                value = str(value)

                key_accessor = re.compile(
                    rf"((\$(?:\:\:)?env\({re.escape(key)}\))([/\-\w\.]*)|(\${re.escape(key)}))"
                )
                for use in key_accessor.findall(script):
                    full = use[0]
                    accessor = use[1] or use[3]
                    env_keys_used.add(key)
                    if verbose:
                        info(f"Found {accessor}…")
                    value_substituted = full.replace(accessor, value)

                    if value_substituted.endswith(".tcl") or value_substituted.endswith(
                        ".sdc"
                    ):
                        if value_substituted not in tcls:
                            tcls.add(value_substituted)
                            tcls_to_process.append(value_substituted)
        except FileNotFoundError:
            warnings.append(f"{current} was not found, might be a product. Skipping…")

        current = shift(tcls_to_process)

    # Phase 4: Copy The Files
    final_env = {}

    def copy(frm, to):
        parents = dirname(to)
        mkdirp(parents)

        def do_copy():
            if isdir(frm):
                shutil.copytree(frm, to)
            else:
                shutil.copyfile(frm, to)

        try:
            incomplete_matches = glob.glob(frm + "*")

            if len(incomplete_matches) == 0:
                raise FileNotFoundError()
            elif len(incomplete_matches) != 1 or incomplete_matches[0] != frm:
                # Prefix For Other Files
                for match in incomplete_matches:
                    if match == frm:
                        # If it's both a file AND a prefix for other files
                        do_copy()
                    else:
                        new_frm = match
                        new_to = to + new_frm[len(frm) :]
                        copy(new_frm, new_to)
            else:
                do_copy()
        except FileNotFoundError:
            warnings.append(f"{frm} was not found, might be a product. Skipping…")
        except Exception as e:
            warnings.append(f"Couldn't copy {frm}: {e}. Skipped.")

    if verbose:
        info("\nProcessing environment variables…\n---")

    ol_root = get_openlane_root()
    loop_guard = 1024
    for key in env_keys_used:
        loop_guard -= 1
        if loop_guard == 0:
            raise Exception("An infinite loop has been detected. Please file an issue.")
        full_value = env[key]
        final_env[key] = ""
        if verbose:
            info(f"Processing {key}: {full_value}")
        for potential_file in full_value.split():
            potential_file_abspath = os.path.abspath(potential_file)
            if potential_file_abspath.startswith(run_path):
                final_env[key] = ""
                relative = relpath(potential_file, run_path)
                final_value = join(".", relative)
                final_path = join(destination_folder, final_value)
                from_path = potential_file
                if potential_file_abspath != step_dir:  # Otherwise, infinite recursion
                    copy(from_path, final_path)
                final_env[key] += f"{final_value} "
            elif potential_file_abspath.startswith(pdk_root):
                if potential_file_abspath == pdk_root:  # Too many files to copy
                    continue
                relative = relpath(potential_file, pdk_root)
                final_value = join("pdk", relative)
                final_path = join(destination_folder, final_value)
                copy(potential_file, final_path)
                final_env[key] += f"{final_value} "
            elif potential_file_abspath.startswith(ol_root):
                relative = relpath(potential_file, ol_root)
                final_value = join("openlane", relative)
                final_path = join(destination_folder, final_value)

                from_path = potential_file
                if potential_file_abspath != os.path.abspath(
                    get_script_dir()
                ):  # Too many files to copy otherwise
                    copy(from_path, final_path)
                final_env[key] += f"{final_value} "
            elif os.path.exists(potential_file) and not potential_file.startswith(
                "/dev"
            ):  # /dev/null, /dev/stdout, /dev/stderr, etc should still work
                final_path = join(destination_folder, potential_file)
                final_value = os.path.relpath(final_path, destination_folder)
                print(potential_file_abspath, final_path)
                copy(potential_file, final_path)
                final_env[key] += f"{final_value} "
            else:
                final_env[key] += f"{potential_file} "
        final_env[key] = final_env[key].rstrip()
    if verbose:
        info("---\n")

    for warning in warnings:
        warn(warning)

    # Phase 5: Create Environment Set/Run Files
    def env_list(
        format_string: str = "{key}={value}",
        env: Dict[str, str] = final_env,
        indent: int = 0,
    ) -> str:
        array = []
        for key, value in sorted(env.items()):
            array.append(format_string.format(key=key, value=value))
        value = f"\n{'    ' * indent}".join(array)
        return value

    run_shell = join(destination_folder, "run.sh")
    with open(run_shell, "w") as f:
        f.write(
            textwrap.dedent(
                f"""\
                #!/bin/sh
                dir=$(cd -P -- "$(dirname -- "$0")" && pwd -P)
                cd $dir;
                {env_list("export {key}='{value}';", indent=4)}
                {' '.join(new_cmd)}
                """
            )
        )
    os.chmod(run_shell, 0o755)

    gdb_env = join(destination_folder, "env.gdb")
    with open(gdb_env, "w") as f:
        f.write(env_list("set env {key} {value}"))

    lldb_env = join(destination_folder, "env.lldb")
    with open(lldb_env, "w") as f:
        f.write(env_list("env {key}={value}"))

    return destination_folder


class TclStep(Step):
    """
    A subclass of :class:`Step` that primarily deals with running Tcl-based utilities,
    such as Yosys, OpenROAD and Magic.

    A TclStep Step corresponds to running one Tcl script with such a utility.
    """

    def get_script_path(self):
        """
        :returns: A Tcl script to be run by this step.
        """
        return os.path.join(get_script_dir(), "tclsh", "hello.tcl")

    def get_command(self) -> List[str]:
        """
        :returns: The command used to run the script.
        """
        return ["tclsh", self.get_script_path()]

    def run(
        self,
        **kwargs,
    ) -> State:
        """
        This overriden :meth:`run` function prepares configuration variables and
        inputs for use with Tcl: specifically, it converts them all to
        environment variables so they may be used by the Tcl scripts being called.

        Additionally, it logs the output to a ``.log`` file named after the script.

        When overriding in a subclass, you may find it useful to use this pattern:

            kwargs, env = self.extract_env(kwargs)
            env["CUSTOM_ENV_VARIABLE"] = "1"
            return super().run(env=env, **kwargs)

        This will allow you to add further custom environment variables to a call
        while still respecting an `env` argument further up the call-stack.


        :param **kwargs: Passed on to subprocess execution: useful if you want to
            redirect stdin, stdout, etc.
        """
        state = super().run()
        command = self.get_command()

        kwargs, env = self.extract_env(kwargs)

        env["SCRIPTS_DIR"] = get_script_dir()
        env["STEP_DIR"] = os.path.abspath(self.step_dir)

        for element in self.config.keys():
            value = self.config[element]
            if value is None:
                continue

            if isinstance(value, list):
                value = " ".join(list(map(lambda x: str(x), value)))
            elif isinstance(value, Enum):
                value = value._name_
            elif isinstance(value, bool):
                value = "1" if value else "0"
            elif isinstance(value, Decimal) or isinstance(value, int):
                value = str(value)

            env[element] = value

        for input in self.inputs:
            if input is None:
                continue
            env[f"CURRENT_{input.name}"] = str(state[input])

        for output in self.outputs:
            filename = f"{self.config['DESIGN_NAME']}.{output.value[1]}"
            env[f"SAVE_{output.name}"] = os.path.join(self.step_dir, filename)

        try:
            self.run_subprocess(
                command,
                env=env,
                **kwargs,
            )
        except subprocess.CalledProcessError:
            reproducible_folder = create_reproducible(
                self.step_dir,
                command,
                env,
                self.get_script_path(),
            )
            info(
                f"Reproducible created: please tarball and upload '{os.path.relpath(reproducible_folder)}' if you're going to file an issue."
            )
            raise

        for output in self.outputs:
            path = Path(env[f"SAVE_{output.name}"])
            if not path.exists():
                continue
            state[output] = path

        return state



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import os
import sys
import tempfile
import subprocess
from typing import List, Tuple

from .flow import Flow
from ..common import get_script_dir
from ..state import State
from ..steps import Step, KLayout


@Flow.factory.register()
class OpenInKLayout(Flow):
    """
    This 'flow' actually just opens the LEF/DEF from the initial state object
    in KLayout. Fancy that.

    Intended for use with run tags that have already been run with
    another flow.
    """

    Steps = [KLayout.StreamOut]

    name = "Opening in KLayout"

    def run(
        self,
        initial_state: State,
        **kwargs,
    ) -> Tuple[State, List[Step]]:
        self.set_max_stage_count(1)
        self.start_stage("Opening in KLayout")

        lefs = ["--input-lef", self.config["TECH_LEF"]]
        for lef in self.config["CELL_LEFS"]:
            lefs.append("--input-lef")
            lefs.append(lef)
        if extra_lefs := self.config["EXTRA_LEFS"]:
            for lef in extra_lefs:
                lefs.append("--input-lef")
                lefs.append(lef)

        subprocess.check_call(
            [
                sys.executable,
                os.path.join(
                    get_script_dir(),
                    "klayout",
                    "open_design.py",
                ),
                "--lyt",
                self.config["KLAYOUT_TECH"],
                "--lyp",
                self.config["KLAYOUT_PROPERTIES"],
                "--lym",
                self.config["KLAYOUT_DEF_LAYER_MAP"],
                initial_state["def"],
            ]
            + lefs
        )
        self.end_stage()

        return (initial_state, [])


@Flow.factory.register()
class OpenInOpenROAD(Flow):
    """
    This 'flow' actually just opens the ODB from the initial state object
    in the OpenROAD GUI.

    Intended for use with run tags that have already been run with
    another flow.
    """

    name = "Opening in OpenROAD"

    Steps = []

    def run(
        self,
        initial_state: State,
        **kwargs,
    ) -> Tuple[State, List[Step]]:
        self.set_max_stage_count(1)

        with tempfile.NamedTemporaryFile("a+", suffix=".tcl") as f:
            f.write(f"read_db \"{initial_state['odb']}\"")
            f.flush()

            subprocess.check_call(
                [
                    "openroad",
                    "-no_splash",
                    "-gui",
                    f.name,
                ]
            )
        self.end_stage()

        return (initial_state, [])



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Optional
from abc import abstractmethod

from .step import Step, StepError, DeferredStepError, State

from ..config import Variable
from ..logging import err, warn, info


class MetricChecker(Step):
    deferred = True

    @abstractmethod
    def get_metric_name(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def get_metric_description(self) -> str:
        raise NotImplementedError()

    def get_threshold(self) -> Optional[Decimal]:
        return Decimal(0)

    def run(self, **kwargs) -> State:
        state_out = super().run()

        metric_name = self.get_metric_name()
        metric_description = self.get_metric_description()
        threshold = self.get_threshold()

        if threshold is None:
            warn(
                f"Threshold for {metric_description} is not set. The checker will be skipped."
            )
        else:
            metric_value = state_out.metrics.get(metric_name)
            if metric_value is not None:
                if metric_value > threshold:
                    error_msg = f"{metric_value} {metric_description} found."
                    if self.deferred:
                        err(f"{error_msg} - deferred")
                        raise DeferredStepError(error_msg)
                    else:
                        err(f"{error_msg}")
                        raise StepError(error_msg)

                else:
                    info(f"Check for {metric_description} clear.")
            else:
                warn(
                    f"The {metric_description} metric was not found. Are you sure the relevant step was run?"
                )

        return state_out


@Step.factory.register()
class YosysUnmappedCells(MetricChecker):
    id = "Checker.YosysUnmappedCells"
    flow_control_variable = "QUIT_ON_UNMAPPED_CELLS"
    name = "Unmapped cells Checker"
    deferred = False

    config_vars = [
        Variable(
            "QUIT_ON_UNMAPPED_CELLS",
            bool,
            "Checks for unmapped cells after sythesis",
            default=True,
        ),
    ]

    def get_metric_name(self) -> str:
        return "design__instance_unmapped__count"

    def get_metric_description(self) -> str:
        return "Unmapped Yosys instances"


@Step.factory.register()
class TrDRC(MetricChecker):
    id = "Checker.TrDRC"
    flow_control_variable = "QUIT_ON_TR_DRC"
    name = "Routing DRC Checker"
    long_name = "Routing Design Rule Check"

    config_vars = [
        Variable(
            "QUIT_ON_TR_DRC",
            bool,
            "Checks for DRC violations after routing and exits the flow if any was found.",
            default=True,
        ),
    ]

    def get_metric_name(self) -> str:
        return "route__drc_errors"

    def get_metric_description(self) -> str:
        return "Routing DRC errors"


@Step.factory.register()
class MagicDRC(MetricChecker):
    id = "Checker.MagicDRC"
    flow_control_variable = "QUIT_ON_MAGIC_DRC"
    name = "Magic DRC Checker"
    long_name = "Magic Design Rule Check"

    config_vars = [
        Variable(
            "QUIT_ON_MAGIC_DRC",
            bool,
            "Checks for DRC violations after magic DRC is executed and exits the flow if any was found.",
            default=True,
        ),
    ]

    def get_metric_name(self) -> str:
        return "magic__drc_errors"

    def get_metric_description(self) -> str:
        return "Magic DRC errors"


@Step.factory.register()
class IllegalOverlap(MetricChecker):
    id = "Checker.IllegalOverlap"
    flow_control_variable = "QUIT_ON_ILLEGAL_OVERLAPS"
    name = "Illegal Overlap Checker"
    long_name = "Spice Extraction-based Illegal Overlap Check"

    config_vars = [
        Variable(
            "QUIT_ON_ILLEGAL_OVERLAPS",
            bool,
            "Checks for illegal overlaps during magic extraction. In some cases, these imply existing undetected shorts in the design. It also exits the flow if any was found.",
            default=True,
        ),
    ]

    def get_metric_name(self) -> str:
        return "magic__illegal__overlaps"

    def get_metric_description(self) -> str:
        return "Magic Illegal Overlap errors"


@Step.factory.register()
class WireLength(MetricChecker):
    id = "Checker.WireLength"
    flow_control_variable = "QUIT_ON_LONG_WIRE"
    name = "Wire Length Threshold Checker"

    metric_name = "route__max__wirelength"
    metric_description = "Threshold-surpassing long wires"

    config_vars = [
        Variable(
            "QUIT_ON_LONG_WIRE",
            bool,
            "Exits the flow if any wire length exceeds the threshold set in the PDK.",
            default=False,
        ),
    ]

    def get_metric_name(self) -> str:
        return "route__max__wirelength"

    def get_metric_description(self) -> str:
        return "Threshold-surpassing long wires"

    def get_threshold(self) -> Optional[Decimal]:
        threshold = self.config["WIRE_LENGTH_THRESHOLD"]
        assert threshold is None or isinstance(threshold, Decimal)
        return threshold


@Step.factory.register()
class LVS(MetricChecker):
    id = "Checker.LVS"
    flow_control_variable = "QUIT_ON_LVS_ERROR"
    name = "LVS Error Checker"
    long_name = "Layout vs. Schematic Error Checker"

    config_vars = [
        Variable(
            "QUIT_ON_LVS_ERROR",
            bool,
            "Checks for LVS errors after netgen LVS is executed and exits the flow if any was found.",
            default=True,
        ),
    ]

    def get_metric_name(self) -> str:
        return "lvs__total__errors"

    def get_metric_description(self) -> str:
        return "LVS errors"



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import subprocess
from typing import List, Tuple, Optional, Type

from .flow import Flow, FlowException, FlowError
from ..state import State
from ..steps import (
    MissingInputError,
    Step,
    StepError,
    DeferredStepError,
)
from ..logging import info, success, err


class SequentialFlow(Flow):
    """
    The simplest Flow, running each Step as a stage, serially,
    with nothing happening in parallel and no significant inter-step
    processing.

    All subclasses of this flow have to do is override the :attr:`.Steps` abstract property
    and it would automatically handle the rest. See `Classic` for an example.
    """

    @classmethod
    def make(Self, step_ids: List[str]) -> Type["SequentialFlow"]:
        step_list = []
        for name in step_ids:
            step = Step.get(name)
            if step is None:
                raise TypeError(f"No step found with id '{name}'")
            step_list.append(step)

        class CustomSequentialFlow(SequentialFlow):
            name = "Custom Sequential Flow"
            Steps = step_list

        return CustomSequentialFlow

    def run(
        self,
        initial_state: State,
        frm: Optional[str] = None,
        to: Optional[str] = None,
        **kwargs,
    ) -> Tuple[State, List[Step]]:
        frm_resolved = None
        to_resolved = None
        for cls in self.Steps:
            if frm is not None and frm.lower() == cls.id.lower():
                frm_resolved = cls.id
            if to is not None and to.lower() == cls.id.lower():
                to_resolved = cls.id

        if frm is not None and frm_resolved is None:
            raise FlowException(f"Failed to process end step '{frm}': step not found")
        if to is not None and to_resolved is None:
            raise FlowException(f"Failed to process end step '{to}': step not found")

        step_count = len(self.Steps)
        self.set_max_stage_count(step_count)

        step_list = []

        info("Starting…")

        executing = frm is None
        deferred_errors = []

        current_state = initial_state
        for cls in self.Steps:
            step = cls(config=self.config, state_in=current_state)
            if frm_resolved is not None and frm_resolved == step.id:
                executing = True

            self.start_stage(step.name)
            if not executing:
                info(f"Skipping step '{step.name}'…")
                self.end_stage(no_increment_ordinal=True)
                continue

            step_list.append(step)
            try:
                current_state = step.start()
            except MissingInputError as e:
                raise FlowException(str(e))
            except DeferredStepError as e:
                deferred_errors.append(str(e))
            except (StepError, subprocess.CalledProcessError) as e:
                raise FlowError(str(e))

            self.end_stage()

            if to_resolved and to_resolved == step.id:
                executing = False
        if len(deferred_errors) != 0:
            err("The following deferred step errors have been encountered:")
            for error in deferred_errors:
                err(error)
            raise FlowError("One or more deferred errors were encountered.")
        success("Flow complete.")
        return (current_state, step_list)



#!/usr/bin/env python3
# Copyright (c) 2021-2022 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Original Copyright Follows
#
# BSD 3-Clause License
#
# Copyright (c) 2018, The Regents of the University of California
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import sys
from typing import TYPE_CHECKING

if "klayout" in os.path.basename(sys.executable):
    import pya
else:
    import click

    @click.command()
    @click.option("-o", "--output", required=True)
    @click.option(
        "-l",
        "--input-lef",
        "input_lefs",
        multiple=True,
    )
    @click.option(
        "-T",
        "--lyt",
        required=True,
        help="KLayout .lyt file",
    )
    @click.option(
        "-P",
        "--lyp",
        required=True,
        help="KLayout .lyp file",
    )
    @click.option(
        "-M",
        "--lym",
        required=True,
        help="KLayout .map (LEF/DEF layer map) file",
    )
    @click.argument("input")
    def cli(**kwargs):
        args = [
            "klayout",
            "-b",
            "-rm",
            __file__,
        ]
        for key, value in kwargs.items():
            args.append("-rd")
            if isinstance(value, tuple) or isinstance(value, list):
                value = ";".join(value)
            elif isinstance(value, str) and os.path.exists(value):
                value = os.path.abspath(value)

            args.append(f"{key}={value or 'NULL'}")

        os.execlp("klayout", *args)

    if __name__ == "__main__":
        cli()


if TYPE_CHECKING:
    # Dummy data for type-checking
    input: str = ""
    output: str = ""
    input_lefs: str = ""
    lyp: str = ""
    lyt: str = ""
    lym: str = ""

try:
    gds = input.endswith(".gds")

    # Load technology file
    tech = pya.Technology()
    tech.load(lyt)

    layout_options = None
    if not gds:
        layout_options = tech.load_layout_options
        layout_options.lefdef_config.map_file = lym
        layout_options.lefdef_config.macro_resolution_mode = 1
        layout_options.lefdef_config.read_lef_with_def = False
        layout_options.lefdef_config.lef_files = input_lefs.split(";")

    view = pya.LayoutView()
    view.load_layer_props(lyp)

    if gds:
        view.load_layout(input)
    else:
        view.load_layout(input, layout_options, lyt)
    view.max_hier()
    pixels = view.get_pixels_with_options(1000, 1000)

    with open(output, "wb") as f:
        f.write(pixels.to_png_data())

    pya.Application.instance().exit(0)
except Exception as e:
    print(e)
    pya.Application.instance().exit(1)



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import json
from decimal import Decimal
from collections import UserDict
from typing import Union, Optional, Dict, Any

from .design_format import DesignFormat, DesignFormatByID

from ..config import Path
from ..logging import warn


class StateEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Decimal):
            if o.as_integer_ratio()[1] == 1:
                return int(o)
            else:
                return float(o)
        elif isinstance(o, Path):
            return str(o)
        return super(StateEncoder, self).default(o)


class StateDecoder(json.JSONDecoder):
    def default(self, o):
        if isinstance(o, float) or isinstance(o, int):
            return Decimal(o)
        return super(StateEncoder, self).default(o)


class State(UserDict):
    """
    Basically, a dictionary with keys of type DesignFormat and string values,
    the string values being filesystem paths.

    The state is the only thing that can be altered by steps other than the
    filesystem.

    This dictionary has a property named `metrics` that also carries statistics
    about the design: area, wire length, et cetera.
    """

    metrics: dict

    def __init__(self, metrics: Optional[dict] = None) -> None:
        super().__init__()
        for format in DesignFormat:
            id: str = format.value[0]
            self[id] = None
        self.metrics = metrics or {}

    def __getitem__(self, key: Union[DesignFormat, str]) -> Optional[Path]:
        if isinstance(key, DesignFormat):
            id: str = key.value[0]
            key = id
        return super().__getitem__(key)

    def __setitem__(self, key: Union[DesignFormat, str], item: Optional[Path]):
        if isinstance(key, DesignFormat):
            id: str = key.value[0]
            key = id
        return super().__setitem__(key, item)

    def as_dict(self) -> dict:
        final: Dict[Any, Any] = dict(self)
        final["metrics"] = self.metrics
        return final

    def __copy__(self: "State") -> "State":
        new = super().__copy__()
        new.metrics = self.metrics.copy()
        return new

    def __repr__(self) -> str:
        return self.as_dict().__repr__()

    def dumps(self, **kwargs) -> str:
        if "indent" not in kwargs:
            kwargs["indent"] = 4
        return json.dumps(self.as_dict(), cls=StateEncoder, **kwargs)

    @classmethod
    def loads(Self, json_in: str, validate_path: bool = True) -> "State":
        raw = json.loads(json_in, cls=StateDecoder)

        metrics = raw.get("metrics")
        if metrics is not None:
            del raw["metrics"]

        state = Self(metrics=metrics)

        for key, value in raw.items():
            df = DesignFormatByID.get(key)
            if df is None:
                warn(f"Unknown design format ID '{key}' in loaded state.")
                continue

            if value is None:
                state[df] = value
                continue

            if validate_path and not os.path.exists(value):
                raise ValueError(
                    f"Provided path '{value}' to design format '{key}' does not exist."
                )
            state[df] = Path(value)

        return state

    def _repr_html_(self) -> str:
        result = """
        <div style="display: grid; grid-auto-columns: minmax(0, 1fr); grid-auto-rows: minmax(0, 1fr); grid-auto-flow: column;">
            <table style="grid-column-start: 1; grid-column-end: 2;">
                <tr>
                    <th>Format</th>
                    <th>Path</th>
                </tr>
        """
        for id, value in dict(self).items():
            assert isinstance(id, str)
            if value is None:
                continue

            format = DesignFormatByID[id]

            value_rel = os.path.relpath(value, ".")

            result += f"""
                <tr>
                    <td>{format.value[2]}</td>
                    <td><a href="{value_rel}">{value_rel}</a></td>
                </tr>
            """

        result += """
            </table>
        </div>
        """

        return result



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
import os
import inspect
from enum import Enum
from typing import get_origin, get_args
from dataclasses import dataclass, field
from decimal import Decimal, InvalidOperation
from typing import Union, Type, List, Optional, Tuple, Any, Dict, Callable, Sequence

from .config import Config, Path
from .resolve import process_string, Keys as SpecialKeys

# Scalar = Union[Type[str], Type[Decimal], Type[Path], Type[bool]]
# VType = Union[Scalar, List[Scalar]]


def StringEnum(name: str, values: Sequence[str]):
    """
    Creates a string enumeration where the keys and values are the same.
    """
    return Enum(name, [(value, value) for value in values])


newline_rx = re.compile(r"\n")


@dataclass
class Variable:
    """
    An object representing a configuration variable for a PDK, a Flow or a Step.

    :param name: A string name for the Variable. Because of backwards compatility
        with OpenLane 1, the convention is ``UPPER_SNAKE_CASE``.

    :param type: A Python type object representing the variable.

        Supported scalars:

        - ``int``
        - ``decimal.Decimal``
        - ``bool``
        - ``str``
        - :class:`Path`
        - Enums

        Supported "generics":

        - ``Union`` (incl. ``Optional``)
        - ``List``
        - ``Tuple``

    :param description: A human-readable description of the variable. Used to
        generate help strings and documentation.

    :param default: A default value for the variable.

        Optional variables have an implicit default value of ``None``.

    :param deprecated_names: A list of deprecated names for said variable.

        An element of the list can alternative be a tuple of a name and a Callable
        used to perform a translation for when a renamed variable is also slightly
        modified.

    :param units: Used only in documentation: the unit corresponding to this
        object, i.e., μm, pF, etc. Can be any string, but for consistency, SI units
        must be represented in terms of their official symbols.
    """

    name: str
    type: Any
    description: str
    default: Any = None
    deprecated_names: List[Union[str, Tuple[str, Callable]]] = field(
        default_factory=list
    )

    units: Optional[str] = None

    def is_optional(self) -> bool:
        """
        Returns whether a variable's type is an `Option type <https://en.wikipedia.org/wiki/Option_type>`_.
        """
        type_args = get_args(self.type)
        return get_origin(self.type) is Union and type(None) in type_args

    def some(self) -> Any:
        """
        Returns the type of a variable presuming it is not None.

        If a variable is not Optional, that is simply the type specified in the
        :attr:`type` field.
        """
        if not self.is_optional():
            return self.type
        else:
            type_args = get_args(self.type)
            if len(type_args) == 2:
                if type_args[0] != Type[None]:
                    return type_args[0]
                else:
                    return type_args[1]
            else:
                some_type_args = list(type_args)
                some_type_args.remove(Type[None])
                return tuple(some_type_args)

    @property
    def required(self) -> bool:
        """
        Whether this variable is required to be specified by a user configuration
        or not.

        What it boils down to is: Variables without defaults that are not optional
        are required.
        """
        return self.default is None and not self.is_optional()

    def type_repr_md(self) -> str:
        """
        Prints a pretty Markdown string representation of the Variable's type.
        """
        some = self.some()
        optional = self.is_optional()

        type_string = some.__name__
        if inspect.isclass(some) and issubclass(some, Enum):
            type_string = "	\\| ".join([repr(e.value) for e in some])
            type_string = f"`{type_string}`"

        origin, args = get_origin(some), get_args(some)
        if origin is not None:
            arg_strings = [arg.__name__ for arg in args]
            if origin == Union:
                type_string = "	\\| ".join(arg_strings)
                type_string = f"({type_string})"
            else:
                type_string = f"{type_string}[{','.join(arg_strings)}]"

        return type_string + ("?" if optional else "")

    def desc_repr_md(self) -> str:
        """
        Prints the description, but with newlines escaped for Markdown.
        """
        return newline_rx.sub("<br />", self.description)

    def _process(
        self,
        exists: bool,
        value: Any,
        validating_type: Optional[Type] = None,
        values_so_far: Optional[Dict[str, Any]] = None,
    ):
        if values_so_far is None:
            values_so_far = {}

        if validating_type is None:
            validating_type = self.type
            if self.is_optional():
                validating_type = self.some()

        assert validating_type is not None

        if not exists:
            if self.default is not None:
                return self._process(True, self.default, validating_type, values_so_far)
            elif self.is_optional():
                return None
            else:
                raise ValueError(
                    f"Required variable {self.name} did not get a specified value."
                )
        elif value is None:
            if self.is_optional():
                return value
            else:
                raise ValueError(
                    f"Required variable {self.name} received a null value."
                )

        if type(value) == str:
            value = process_string(value, values_so_far)

        if get_origin(validating_type) == list:
            return_value = list()
            subtype = get_args(validating_type)[0]
            if isinstance(value, list):
                pass
            elif isinstance(value, str):
                if "," in value:
                    value = value.split(",")
                elif ";" in value:
                    value = value.split(";")
                else:
                    value = value.split()
            else:
                raise ValueError(
                    f"Invalid List provided for variable {self.name}: {value}"
                )
            for item in value:
                return_value.append(self._process(True, item, subtype))
            return return_value
        elif validating_type == Path:
            if not os.path.exists(value):
                raise ValueError(
                    f"Path provided for variable {self.name} does not exist: '{value}'"
                )
            return Path(value)
        elif validating_type == bool:
            if value in ["1", "true", True]:
                return True
            elif value in ["0", "false", False]:
                return False
            else:
                raise ValueError(
                    f"Value provided for variable {self.name} of type {validating_type} is invalid: '{value}'"
                )
        elif issubclass(validating_type, Enum):
            try:
                return validating_type[value]
            except KeyError:
                raise ValueError(
                    f"Variable provided for variable {self.name} of enumerated type {validating_type} is invalid: 'value'"
                )
        elif issubclass(validating_type, Decimal):
            try:
                return Decimal(value)
            except InvalidOperation:
                raise ValueError(
                    f"Value provided for variable {self.name} of type {validating_type} is invalid: '{value}'"
                )
        else:
            try:
                return validating_type(value)
            except ValueError as e:
                raise ValueError(
                    f"Value provided for variable {self.name} of type {validating_type} is invalid: '{value}' {e}"
                )

    def _compile(
        self,
        mutable_config: Config,
        warning_list_ref: List[str],
        values_so_far: Config,
    ) -> Any:
        exists, value = mutable_config.extract(self.name)

        i = 0
        while (
            not exists
            and self.deprecated_names is not None
            and i < len(self.deprecated_names)
        ):
            deprecated_name = self.deprecated_names[i]
            deprecated_callable = lambda x: x
            if not isinstance(deprecated_name, str):
                deprecated_name, deprecated_callable = deprecated_name
            exists, value = mutable_config.extract(deprecated_name)
            if exists:
                warning_list_ref.append(
                    f"The configuration variable '{deprecated_name}' is deprecated. Please check the docs for the usage on the replacement variable '{self.name}'."
                )
            if value is not None:
                value = deprecated_callable(value)
            i = i + 1

        return self._process(exists, value, values_so_far=values_so_far.data)

    def __hash__(self) -> int:
        return hash((self.name, self.description))

    def __eq__(self, rhs: object) -> bool:
        if not isinstance(rhs, Variable):
            raise NotImplementedError()
        return (
            self.name == rhs.name
            and self.description == rhs.description
            and self.type == rhs.type
            and self.default == rhs.default
        )

    @classmethod
    def process_config(
        Self,
        config: Config,
        variables: Sequence["Variable"],
        removed: Optional[Dict[str, str]] = None,
    ) -> Tuple[Config, List[str], List[str]]:
        """
        Verifies a configuration object against a list of variables, returning
        an object with the variables normalized according to their types.

        :param config: The input, raw configuration object.
        :param variables: A sequence or some other iterable of variables.
        :param removed: A dictionary of variables that may have existed at a point in
            time, but then have gotten removed. Useful to give feedback to the user.
        :returns: A tuple of:
            [0] A final, processed configuration.
            [1] A list of warnings.
            [2] A list of errors.

            If the third element is non-empty, the first object is invalid.
        """
        if removed is None:
            removed = {}

        warnings: List[str] = []
        errors = []
        final = Config()
        mutable = config.copy()
        for variable in variables:
            try:
                value_processed = variable._compile(
                    mutable_config=mutable,
                    warning_list_ref=warnings,
                    values_so_far=final,
                )
                final[variable.name] = value_processed
            except ValueError as e:
                errors.append(str(e))

        for key in sorted(mutable.keys()):
            if key in vars(SpecialKeys).values():
                continue
            if key in removed:
                warnings.append(f"'{key}' has been removed: {removed[key]}")
            elif "_OPT" not in key:
                warnings.append(f"Unknown key {key} provided.")

        return (final, warnings, errors)



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import atexit
import click
import typing
import logging

import rich.console
from rich.logging import RichHandler


console = rich.console.Console()
atexit.register(lambda: rich.console.Console().show_cursor())


class LogLevels:
    ALL: typing.ClassVar[int] = 0
    DEBUG: typing.ClassVar[int] = 10
    VERBOSE: typing.ClassVar[int] = 15
    INFO: typing.ClassVar[int] = 20
    WARN: typing.ClassVar[int] = 30
    ERROR: typing.ClassVar[int] = 40
    CRITICAL: typing.ClassVar[int] = 50


LogLevelsDict = {k: v for k, v in LogLevels.__dict__.items() if not k.startswith("__")}

for name, value in LogLevelsDict.items():
    logging.addLevelName(value, name)

FORMAT = "%(message)s"
logging.basicConfig(
    level=15,
    format=FORMAT,
    datefmt="",
    handlers=[
        RichHandler(
            console=console,
            rich_tracebacks=True,
            markup=True,
            tracebacks_suppress=[
                click,
            ],
        )
    ],
)


def set_log_level(lv: typing.Union[str, int]):
    logging.getLogger().setLevel(lv)


def get_log_level() -> int:
    return logging.getLogger().getEffectiveLevel()


def print(*args, **kwargs):
    """
    Prints to the terminal without formatting.

    See the Rich API for more info: https://rich.readthedocs.io/en/stable/reference/console.html#rich.console.Console.print
    """
    if get_log_level() > LogLevels.INFO:
        return
    console.print(*args, **kwargs)


def verbose(*args, **kwargs):
    """
    Prints to the terminal with automatic rich formatting.

    Verbose level or lower only.

    See the Rich API for more info: https://rich.readthedocs.io/en/stable/reference/console.html#rich.console.Console.log
    """
    if get_log_level() > LogLevels.VERBOSE:
        return
    console.print(*args, **kwargs)


def info(*args, **kwargs):
    """
    Prints to the terminal with automatic rich formatting.

    See the Rich API for more info: https://rich.readthedocs.io/en/stable/reference/console.html#rich.console.Console.log
    """
    if get_log_level() > LogLevels.INFO:
        return
    if kwargs.get("_stack_offset") is None:
        kwargs["_stack_offset"] = 2
    console.log(*args, **kwargs)


def rule(*args, **kwargs):
    """
    Prints a horizontal line on the terminal.

    See the Rich API for more info: https://rich.readthedocs.io/en/stable/reference/console.html#rich.console.Console.rule
    """
    if get_log_level() > LogLevels.INFO:
        return
    console.rule(*args, **kwargs)


def success(printable, *args, **kwargs):
    """
    Logs an item to the terminal with a success unicode character and
    green/bold formatting.
    """
    if get_log_level() > LogLevels.INFO:
        return
    if kwargs.get("_stack_offset") is None:
        kwargs["_stack_offset"] = 2
    console.log(f"⭕ [green][bold] {printable}", *args, **kwargs)


def warn(printable, *args, **kwargs):
    """
    Logs an item to the terminal with a warning unicode character and
    gold/bold formatting.
    """
    if get_log_level() > LogLevels.WARN:
        return
    if kwargs.get("_stack_offset") is None:
        kwargs["_stack_offset"] = 2
    console.log(f"⚠️ [gold][bold] {printable}", *args, **kwargs)


def err(printable, *args, **kwargs):
    """
    Logs an item to the terminal with a warning unicode character and
    gold/bold formatting.
    """
    if get_log_level() > LogLevels.ERROR:
        return
    if kwargs.get("_stack_offset") is None:
        kwargs["_stack_offset"] = 2
    console.log(f"❌ [red][bold] {printable}", *args, **kwargs)



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
from typing import Optional
from base64 import b64encode

from .step import Step, StepError
from ..state import DesignFormat, State

from ..logging import warn
from ..config import Path, Variable
from ..common import get_script_dir


@Step.factory.register()
class StreamOut(Step):
    id = "KLayout.StreamOut"
    name = "GDS-II Stream Out (KLayout)"
    flow_control_variable = "RUN_KLAYOUT_STREAMOUT"

    inputs = [DesignFormat.DEF]
    outputs = [DesignFormat.GDS, DesignFormat.KLAYOUT_GDS]

    config_vars = [
        Variable(
            "RUN_KLAYOUT_STREAMOUT",
            bool,
            "Enables streaming GDS-II using KLayout.",
            default=True,
            deprecated_names=["RUN_KLAYOUT"],
        )
    ]

    def run(self, **kwargs) -> State:
        state_out = super().run(**kwargs)

        assert isinstance(self.state_in, State)

        lyp = self.config["KLAYOUT_PROPERTIES"]
        lyt = self.config["KLAYOUT_TECH"]
        lym = self.config["KLAYOUT_DEF_LAYER_MAP"]
        if None in [lyp, lyt, lym]:
            if self.config["PRIMARY_SIGNOFF_TOOL"].value == "klayout":
                raise StepError(
                    "One of KLAYOUT_PROPERTIES, KLAYOUT_TECH or KLAYOUT_DEF_LAYER_MAP is unset, yet, KLayout is set as the primary sign-off tool."
                )
            warn(
                "One of KLAYOUT_PROPERTIES, KLAYOUT_TECH or KLAYOUT_DEF_LAYER_MAP is unset. Skipping KLayout stream-out…"
            )
            return state_out

        klayout_gds_out = os.path.join(
            self.step_dir,
            f"{self.config['DESIGN_NAME']}.{DesignFormat.KLAYOUT_GDS.value[1]}",
        )

        layout_args = [
            "--input-lef",
            self.config["TECH_LEF"],
        ]
        for lef in self.config["CELL_LEFS"]:
            layout_args.append("--input-lef")
            layout_args.append(lef)
        if extra_lefs := self.config["EXTRA_LEFS"]:
            for lef in extra_lefs:
                layout_args.append("--input-lef")
                layout_args.append(lef)
        for gds in self.config["CELL_GDS"]:
            layout_args.append("--with-gds-file")
            layout_args.append(gds)
        if extra_gds := self.config["EXTRA_GDS_FILES"]:
            for gds in extra_gds:
                layout_args.append("--with-gds-file")
                layout_args.append(gds)

        kwargs, env = self.extract_env(kwargs)

        self.run_subprocess(
            [
                sys.executable,
                os.path.join(
                    get_script_dir(),
                    "klayout",
                    "stream_out.py",
                ),
                self.state_in[DesignFormat.DEF.value[0]],
                "--output",
                klayout_gds_out,
                "--lyt",
                lyt,
                "--lyp",
                lyp,
                "--lym",
                lym,
                "--top",
                self.config["DESIGN_NAME"],
            ]
            + layout_args,
            env=env,
        )

        state_out[DesignFormat.KLAYOUT_GDS] = Path(klayout_gds_out)

        if self.config["PRIMARY_SIGNOFF_TOOL"].value == "klayout":
            state_out[DesignFormat.GDS] = state_out[DesignFormat.KLAYOUT_GDS]

        return state_out

    def layout_preview(self) -> Optional[str]:
        if self.state_out is None:
            return None
        assert self.toolbox is not None

        if image := self.toolbox.render_png(
            self.config, str(self.state_out["klayout_gds"])
        ):
            image_encoded = b64encode(image).decode("utf8")
            return f'<img src="data:image/png;base64,{image_encoded}" />'

        return None


@Step.factory.register()
class XOR(Step):
    id = "KLayout.XOR"
    name = "KLayout vs. Magic XOR"
    flow_control_variable = "RUN_KLAYOUT_XOR"

    inputs = [
        DesignFormat.MAG_GDS,
        DesignFormat.KLAYOUT_GDS,
    ]

    config_vars = [
        Variable(
            "RUN_KLAYOUT_XOR",
            bool,
            "Enables running KLayout XOR on the two GDS-II files generated by Magic and Klayout. Stream-outs for both KLayout and Magic should have already run, and the PDK must support both signoff tools.",
            default=True,
        ),
        Variable(
            "KLAYOUT_XOR_THREADS",
            int,
            "Specifies number of threads used in the KLayout XOR check.",
            default=1,
        ),
    ]

    def run(self, **kwargs) -> State:
        state_out = super().run(**kwargs)

        ignored = ""
        if ignore_list := self.config["KLAYOUT_XOR_IGNORE_LAYERS"]:
            ignored = ";".join(ignore_list)

        layout_a = state_out[DesignFormat.MAG_GDS]
        if layout_a is None:
            warn("No Magic stream-out has been performed. Skipping XOR…")
            return state_out
        layout_b = state_out[DesignFormat.KLAYOUT_GDS]
        if layout_b is None:
            warn("No KLayout stream-out has been performed. Skipping XOR…")
            return state_out

        kwargs, env = self.extract_env(kwargs)

        self.run_subprocess(
            [
                "ruby",
                os.path.join(
                    get_script_dir(),
                    "klayout",
                    "xor.drc",
                ),
                "--output",
                os.path.join(self.step_dir, "xor.xml"),
                "--top",
                self.config["DESIGN_NAME"],
                "--ignore",
                ignored,
                layout_a,
                layout_b,
            ],
            env=env,
        )

        return state_out



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
import os
import glob
import fnmatch
from enum import Enum
from decimal import Decimal
from types import SimpleNamespace
from typing import Any, Dict, List, Tuple, Union, Optional

Keys = SimpleNamespace(
    pdk_root="PDK_ROOT",
    pdk="PDK",
    pdkpath="PDKPATH",
    scl="STD_CELL_LIBRARY",
    design_dir="DESIGN_DIR",
)

PROCESS_INFO_ALLOWLIST = [
    Keys.pdk,
    Keys.scl,
    f"{Keys.scl}_OPT",
]


Scalar = Union[str, int, float, bool, None]


class InvalidConfig(Exception):
    pass


State = dict


class Expr(object):
    class SyntaxError(Exception):
        pass

    class Token(object):
        class Type(Enum):
            VAR = 0
            NUMBER = 1
            OP = 2
            LPAREN = 3
            RPAREN = 4

        def __init__(self, type: "Expr.Token.Type", value: str) -> None:
            self.type: Expr.Token.Type = type
            self.value: str = value

        def __repr__(self):
            return f"<Token:{self.type} '{self.value}'>"

        def prec_assoc(self) -> Tuple[int, bool]:
            """
            Returns (precedence, is_left_assoc)
            """

            if self.value in ["**"]:
                return (20, False)
            elif self.value in ["*", "/"]:
                return (10, True)
            elif self.value in ["+", "-"]:
                return (0, True)
            else:
                raise TypeError(
                    f"pre-assoc not supported for non-token operators: '{self.value}'"
                )

    @staticmethod
    def tokenize(expr: str) -> List["Expr.Token"]:
        rx_list = [
            (re.compile(r"^\$(\w+)"), Expr.Token.Type.VAR),
            (re.compile(r"^(-?\d+\.?\d*)"), Expr.Token.Type.NUMBER),
            (re.compile(r"^(\*\*)"), Expr.Token.Type.OP),
            (re.compile(r"^(\+|\-|\*|\/)"), Expr.Token.Type.OP),
            (re.compile(r"^(\()"), Expr.Token.Type.LPAREN),
            (re.compile(r"^(\))"), Expr.Token.Type.RPAREN),
            (re.compile(r"^\s+"), None),
        ]
        tokens = []
        str_so_far = expr
        while not str_so_far.strip() == "":
            found = False

            for element in rx_list:
                rx, type = element
                m = rx.match(str_so_far)
                if m is None:
                    continue
                found = True
                if type is not None:
                    tokens.append(Expr.Token(type, m[1]))
                str_so_far = str_so_far[len(m[0]) :]
                break

            if not found:
                raise SyntaxError(
                    f"Unexpected token at the start of the following string '{str_so_far}'."
                )
        return tokens

    @staticmethod
    def evaluate(expression: str, vars: Dict[str, str]) -> float:
        tokens: List["Expr.Token"] = Expr.tokenize(expression)
        ETT = Expr.Token.Type

        # Infix to Postfix
        postfix: List["Expr.Token"] = []
        opstack: List["Expr.Token"] = []
        for token in tokens:
            if token.type == ETT.OP:
                prec, assoc = token.prec_assoc()

                top_prec = None
                try:
                    top_prec, _ = opstack[-1].prec_assoc()
                except TypeError:
                    pass
                except IndexError:
                    pass

                while top_prec is not None and (
                    (assoc and prec <= top_prec) or (not assoc and prec < top_prec)
                ):
                    postfix.append(opstack.pop())
                    top_prec = None
                    try:
                        top_prec, _ = opstack[-1].prec_assoc()
                    except IndexError:
                        pass
                opstack.append(token)
            elif token.type == ETT.LPAREN:
                opstack.append(token)
            elif token.type == ETT.RPAREN:
                top = opstack[-1]
                while top.type != ETT.LPAREN:
                    postfix.append(top)
                    opstack.pop()
                    top = opstack[-1]
                opstack.pop()  # drop the LPAREN
            else:
                postfix.append(token)

        while len(opstack):
            postfix.append(opstack[-1])
            opstack.pop()

        # Evaluate
        eval_stack = []
        for token in postfix:
            if token.type == ETT.NUMBER:
                eval_stack.append(float(token.value))
            elif token.type == ETT.VAR:
                try:
                    value = vars[token.value]
                    eval_stack.append(float(value))
                except KeyError:
                    raise SyntaxError(
                        f"Configuration variable '{token.value}' not found."
                    )
                except Exception:
                    raise SyntaxError(
                        f"Invalid non-numeric value '{value}' for variable ${token.value}."
                    )
            elif token.type == ETT.OP:
                try:
                    number1 = eval_stack[-2]
                    number2 = eval_stack[-1]
                    eval_stack.pop()
                    eval_stack.pop()

                    result = 0.0
                    if token.value == "**":
                        result = number1**number2
                    elif token.value == "*":
                        result = number1 * number2
                    elif token.value == "/":
                        result = number1 / number2
                    elif token.value == "+":
                        result = number1 + number2
                    elif token.value == "-":
                        result = number1 + number2

                    eval_stack.append(result)
                except IndexError:
                    raise SyntaxError(
                        f"Not enough operands for operator '{token.value}'."
                    )

        if len(eval_stack) > 1:
            raise SyntaxError("Expression does not reduce to one value.")
        elif len(eval_stack) == 0:
            raise SyntaxError("Expression is empty.")

        return eval_stack[0]


ref_rx = re.compile(r"^\$([A-Za-z_][A-Za-z0-9_]*)")


def process_string(value: str, state: State) -> Optional[str]:
    global ref_rx
    EXPR_PREFIX = "expr::"
    REF_PREFIX = "ref::"

    DIR_PREFIX = "dir::"
    PDK_DIR_PREFIX = "pdk_dir::"

    if value.startswith(DIR_PREFIX):
        value = value.replace(DIR_PREFIX, f"ref::${Keys.design_dir}/")
    elif value.startswith(PDK_DIR_PREFIX):
        value = value.replace(PDK_DIR_PREFIX, f"ref::${Keys.pdkpath}/")

    if value.startswith(EXPR_PREFIX):
        try:
            value = f"{Expr.evaluate(value[len(EXPR_PREFIX):], state)}"
        except SyntaxError as e:
            raise InvalidConfig(f"Invalid expression '{value}': {e}")
    elif value.startswith(REF_PREFIX):
        reference = value[len(REF_PREFIX) :]
        match = ref_rx.match(reference)
        if match is None:
            raise InvalidConfig(f"Invalid reference string '{reference}'")
        reference_variable = match[1]
        try:
            found = state[reference_variable]
            if found is None:
                return None

            if type(found) != str:
                if type(found) in [int, float]:
                    raise InvalidConfig(
                        f"Referenced variable {reference_variable} is a number and not a string: use expr::{match[0]} if you want to reference this number."
                    )
                else:
                    raise InvalidConfig(
                        f"Referenced variable {reference_variable} is not a string: {type(found)}."
                    )

            value = reference.replace(match[0], found)

            found_abs = os.path.abspath(found)
            full_abspath = os.path.abspath(value)

            # Resolve globs for paths that are inside the exposed directory
            if full_abspath.startswith(found_abs):
                files = glob.glob(full_abspath)
                files_escaped = [file.replace("$", r"\$") for file in files]
                if len(files_escaped) != 0:
                    value = " ".join(files_escaped)
        except KeyError:
            raise InvalidConfig(
                f"Referenced variable '{reference_variable}' not found."
            )
    return value


def process_scalar(key: str, value: Scalar, state: State) -> Scalar:
    if isinstance(value, str):
        value = process_string(value, state)
    elif value is None:
        value = ""
    elif not (
        isinstance(value, bool)
        or isinstance(value, int)
        or isinstance(value, float)
        or isinstance(value, Decimal)
    ):
        raise InvalidConfig(f"Invalid value type {type(value)} for key '{key}'.")

    return value


def process_config_dict_recursive(config_in: Dict[str, Any], state: State):
    PDK_PREFIX = "pdk::"
    SCL_PREFIX = "scl::"

    for key, value in config_in.items():
        withhold = False
        if not isinstance(key, str):
            raise InvalidConfig(f"Invalid key {key}: must be a string.")
        if isinstance(value, dict):
            withhold = True
            if key.startswith(PDK_PREFIX):
                pdk_match = key[len(PDK_PREFIX) :]
                if fnmatch.fnmatch(state[Keys.pdk], pdk_match):
                    process_config_dict_recursive(value, state)
            elif key.startswith(SCL_PREFIX):
                scl_match = key[len(SCL_PREFIX) :]
                if state[Keys.scl] is not None and fnmatch.fnmatch(
                    state[Keys.scl], scl_match
                ):
                    process_config_dict_recursive(value, state)
            else:
                raise InvalidConfig(
                    f"Invalid value type {type(value)} for key '{key}'."
                )
        elif isinstance(value, list):
            valid = True
            processed = []
            for i, item in enumerate(value):
                current_key = f"{key}[{i}]"
                processed.append(f"{process_scalar(current_key, item, state)}")

            if not valid:
                raise InvalidConfig(
                    f"Invalid value for key '{key}': Arrays must consist only of strings."
                )
            value = " ".join(processed)
        else:
            value = process_scalar(key, value, state)

        if not withhold:
            state[key] = value


def process_config_dict(config_in: dict, exposed_variables: Dict[str, str]):
    state = State(exposed_variables)
    process_config_dict_recursive(config_in, state)
    return state


def extract_process_vars(config_in: Dict[str, str]) -> Dict[str, str]:
    return {
        key: config_in[key]
        for key in PROCESS_INFO_ALLOWLIST
        if config_in.get(key) is not None and config_in.get(key) != ""
    }


def resolve(
    config_dict: Dict[str, Any],
    design_dir: str,
    only_extract_process_info: bool = False,
    exposing: Optional[List[str]] = None,
    pdk: Optional[str] = None,
    pdkpath: Optional[str] = None,
    scl: Optional[str] = None,
):
    if exposing is None:
        exposing = []

    exposed_dict = {}

    implicitly_exposed = [
        Keys.pdk,
        Keys.pdkpath,
        Keys.scl,
        Keys.design_dir,
    ]
    if only_extract_process_info:
        implicitly_exposed = [Keys.design_dir]

        # So we don't crash on conditional processing:
        def coalesce(variable_name: str):
            nonlocal implicitly_exposed, exposed_dict
            if os.environ.get(variable_name) is not None:
                implicitly_exposed += [variable_name]
            else:
                exposed_dict[variable_name] = ""

        for variable in [
            Keys.pdk,
            Keys.pdkpath,
            Keys.scl,
        ]:
            coalesce(variable)

    base_vars = {
        Keys.pdk: pdk,
        Keys.pdkpath: pdkpath,
        Keys.scl: scl,
        Keys.design_dir: design_dir,
    }
    base_vars_clean = {k: v for k, v in base_vars.items() if v is not None}

    env_copy = os.environ.copy()
    env_copy.update(base_vars_clean)

    exposed = list(exposing) + implicitly_exposed
    for key in exposed:
        if value := env_copy.get(key):
            exposed_dict[key] = value

    for key in implicitly_exposed:
        if exposed_dict.get(key) is None:
            raise ValueError(
                f"{key} environment variable must be set.",
            )

    resolved = process_config_dict(config_dict, exposed_dict)
    if only_extract_process_info:
        resolved = extract_process_vars(resolved)

    resolved.update(base_vars_clean)
    return resolved



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import json
import yaml
from enum import Enum
from decimal import Decimal
from textwrap import dedent
from dataclasses import dataclass, asdict
from collections import UserDict, UserString
from typing import (
    Any,
    ClassVar,
    Tuple,
    Union,
    List,
    Optional,
)


class Path(UserString, os.PathLike):
    """
    A Path type for OpenLane configuration variables.

    Basically just a string.
    """

    def __fspath__(self) -> str:
        return str(self)

    def exists(self) -> bool:
        """
        A convenience method calling :meth:`os.path.exists`
        """
        return os.path.exists(self)


@dataclass
class Meta:
    version: int = 1
    flow: Union[None, str, List[str]] = "Classic"


class ConfigEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Meta):
            return asdict(o)
        elif isinstance(o, Decimal):
            if o.as_integer_ratio()[1] == 1:
                return int(o)
            else:
                return float(o)
        elif isinstance(o, Path):
            return str(o)
        elif isinstance(o, Enum):
            return str(o)
        return super(ConfigEncoder, self).default(o)


class Config(UserDict):
    """
    A map from OpenLane configuration variable keys to their values.

    It is recommended that you construct these using the :class:`ConfigBuilder`
    singleton class.
    """

    current_interactive: ClassVar[Optional["Config"]] = None

    meta: Meta = Meta(version=1)
    interactive: bool = False

    @classmethod
    def get_meta(
        Self,
        json_config_in: Union[str, os.PathLike],
        flow_override: Optional[str] = None,
    ) -> Optional[Meta]:
        """
        Returns the Meta object of a JSON configuration file

        :param config_in: A configuration file.
        :returns: Either a Meta object, or if the file is not a JSON file, None.
        """
        try:
            obj = json.load(open(json_config_in, encoding="utf8"))
        except (json.JSONDecodeError, IsADirectoryError):
            return None

        meta = Meta()
        if meta_raw := obj.get("meta"):
            meta = Meta(**meta_raw)

        if flow_override is not None:
            meta.flow = flow_override

        return meta

    def dumps(self) -> str:
        """
        :returns: A JSON string representing the configuration object.
        """
        data = self.data.copy()
        data["meta"] = self.meta
        return json.dumps(data, indent=2, cls=ConfigEncoder, sort_keys=True)

    def check(self, key: str) -> Tuple[bool, Any]:
        """
        :param key: A string key representing an OpenLane configuration variable.
        :returns: A tuple of whether that key exists, and if so, its value.

            Do note ``None`` is a valid value, so simply checking if the
            second element ``is not None`` is invalid to check whether
            a key exists.
        """
        return (key in self.keys(), self.get(key))

    def extract(self, key: str) -> Tuple[bool, Any]:
        """
        A mutating method that attempts to find a key, and, if it exists,
        deletes it from the configuration object and returns its value.

        :param key: A string key representing an OpenLane configuration variable.
        :returns: A tuple of whether that key existed, and if so its value.

            Do note ``None`` is a valid value, so simply checking if the
            second element ``is not None`` is invalid to check whether
            a key exists.
        """
        found, value = self.check(key)
        if found:
            del self[key]
        return (found, value)

    def _repr_markdown_(self) -> str:
        title = "Interactive Configuration" if self.interactive else "Configuration"
        values_title = "Initial Values" if self.interactive else "Values"
        return (
            dedent(
                f"""
                ### {title}
                #### {values_title}

                <br />

                ```yml
                %s
                ```
                """
            )
            % yaml.safe_dump(json.loads(self.dumps()))
        )



#!/usr/bin/env python3
# Copyright 2020 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import odb

import os
from subprocess import Popen, PIPE

from reader import OdbReader, click


def invoke_padring(config_file_name, output_file_name, lefs):
    print("Invoking padring to generate a padframe")
    padring_command = []
    padring_command.append("padring")
    for lef in lefs:
        padring_command.extend(["-L", lef])
    padring_command.extend(["--def", output_file_name])
    padring_command.append(config_file_name)

    p = Popen(padring_command, stdout=PIPE, stdin=PIPE, stderr=PIPE, encoding="utf8")

    output = p.communicate()
    print("STDERR:")
    print("\n".join(output[1].splitlines()[-10:]))
    print("STDOUT:")
    print(output[0].strip())

    print("Padring exit code:", p.returncode)
    assert p.returncode == 0, p.returncode
    assert os.path.exists(output_file_name)


def chunker(seq, size):
    chunks = [seq[i::size] for i in range(size)]
    # sort by type
    chunks.sort(key=lambda pad_pair: pad_pair[1])
    return chunks


def diff_lists(l1, l2):
    return list(list(set(l1) - set(l2)) + list(set(l2) - set(l1)))


def generate_cfg(north, east, south, west, corner_pads, width, height):
    cfg = []
    cfg.append(f"AREA {width} {height} ;")
    cfg.append("")

    assert len(corner_pads) == 4, corner_pads
    cfg.append(f"CORNER {corner_pads[0][0]} SW {corner_pads[0][1]} ;")
    cfg.append(f"CORNER {corner_pads[1][0]} NW {corner_pads[1][1]} ;")
    cfg.append(f"CORNER {corner_pads[2][0]} NE {corner_pads[2][1]} ;")
    cfg.append(f"CORNER {corner_pads[3][0]} SE {corner_pads[3][1]} ;")

    cfg.append("")

    for pad in north:
        cfg.append(f"PAD {pad[0]} N {pad[1]} ;")

    cfg.append("")

    for pad in east:
        cfg.append(f"PAD {pad[0]} E {pad[1]} ;")

    cfg.append("")

    for pad in south:
        cfg.append(f"PAD {pad[0]} S {pad[1]} ;")

    cfg.append("")

    for pad in west:
        cfg.append(f"PAD {pad[0]} W {pad[1]} ;")

    return "\n".join(cfg)


@click.command()
@click.option(
    "--output",
    required=True,
    help="Output ODB file",
)
@click.option(
    "--output-def",
    help="Output DEF file",
)
@click.option(
    "-v",
    "--verilog-netlist",
    default=None,
    help="A verilog netlist containing pads and other user macros",
)
@click.option(
    "-d",
    "--def-netlist",
    default=None,
    help="A DEF netlist containing unplaced pads and other user macros",
)
@click.option(
    "-l",
    "--input-lef",
    "input_lefs",
    multiple=True,
    required=True,
    help="LEF file(s) needed to have a proper view of the DEF files",
)
@click.option(
    "--odb-lef",
    required=True,
    help="LEF file to be included in output odb",
)
@click.option("-w", "--width", help="Width of the die area.")
@click.option("-h", "--height", help="Height of the die area.")
@click.option(
    "-c",
    "--padframe-config",
    required=True,
    help="Configuration file: input to padring",
)
@click.option(
    "-P",
    "--pad-name-prefixes",
    multiple=True,
    help="Padname prefixes",
)
@click.option(
    "-i",
    "--init-only",
    default=False,
    is_flag=True,
    help="Only generate a configuration file to be user edited.",
)
@click.option(
    "-C",
    "--working-dir",
    default=".",
    help="Working directory to create temporary files needed.",
)
@click.option(
    "-s", "--special-nets", multiple=True, help="Net names to mark as special"
)
@click.argument("design")
def padringer(
    output,
    output_def,
    verilog_netlist,
    def_netlist,
    input_lefs,
    odb_lef,
    width,
    height,
    padframe_config,
    pad_name_prefixes,
    init_only,
    working_dir,
    special_nets,
    design,
):
    """
    Reads in a structural verilog containing pads and a LEF file that
    contains at least those pads and produces a DEF file with the padframe.
    TODO:
    core placement
    config init,
    external config
    """

    config_file_name = padframe_config
    lefs = input_lefs

    working_def = f"{working_dir}/{design}.pf.def"
    working_cfg = f"{working_dir}/{design}.pf.cfg"

    for lef in lefs:
        assert os.path.exists(lef), lef + " doesn't exist"

    # hard requirement of a user netlist either as a DEF or verilog
    # this is to ensure that the padframe will contain all pads in the design
    # whether the config is autogenerated or user-provided
    assert (
        verilog_netlist is not None or def_netlist is not None
    ), "One of --verilog_netlist or --def-netlist is required"

    # Step 1: create an openDB database from the verilog/def using OpenSTA's read_verilog
    if verilog_netlist is not None:
        assert (
            def_netlist is None
        ), "Only one of --verilog_netlist or --def-netlist is required"
        assert design is not None, "--design is required"

        openroad_script = []
        for lef in lefs:
            openroad_script.append(f"read_lef {lef}")
        openroad_script.append(f"read_verilog {verilog_netlist}")
        openroad_script.append(f"link_design {design}")
        openroad_script.append(f"write_def {working_def}")
        openroad_script.append("exit")

        p = Popen(["openroad"], stdout=PIPE, stdin=PIPE, stderr=PIPE, encoding="utf8")

        openroad_script = "\n".join(openroad_script)

        output = p.communicate(openroad_script)
        print("STDOUT:")
        print(output[0].strip())
        print("STDERR:")
        print(output[1].strip())
        print("openroad exit code:", p.returncode)
        assert p.returncode == 0, p.returncode
        # TODO: check for errors
    else:
        assert def_netlist is not None
        working_def = def_netlist

    assert os.path.exists(working_def), "DEF file doesn't exist"

    top = OdbReader(odb_lef, working_def)

    print(f"Top-level design name: {top.name}")

    ## Step 2: create a simple data structure with pads from the library
    # types: corner, power, other
    pads = {}
    libs = top.db.getLibs()
    for lib in libs:
        masters = lib.getMasters()
        for m in masters:
            name = m.getName()
            if m.isPad():
                assert any(name.startswith(p) for p in pad_name_prefixes), name
                print("Found pad:", name)
                pad_type = m.getType()
                pads[name] = pad_type
                if pad_type == "PAD_SPACER":
                    print("Found PAD_SPACER:", name)
                elif pad_type == "PAD_AREAIO":
                    # using this for special bus fillers...
                    print("Found PAD_AREAIO", name)
            if m.isEndCap():
                # FIXME: regular endcaps
                assert any(name.startswith(p) for p in pad_name_prefixes), name
                assert not m.isPad(), name + " is both pad and endcap?"
                print("Found corner pad:", name)
                pads[name] = "corner"

    print()
    print("The I/O library contains", len(pads), "cells")
    print()

    assert len(pads) != 0, "^"

    ## Step 3: Go over instances in the design and extract the used pads
    used_pads = []
    used_corner_pads = []
    other_instances = []
    for inst in top.block.getInsts():
        inst_name = inst.getName()
        master_name = inst.getMaster().getName()
        if inst.isPad():
            assert any(
                master_name.startswith(p) for p in pad_name_prefixes
            ), master_name
            print("Found pad instance", inst_name, "of type", master_name)
            used_pads.append((inst_name, master_name))
        elif inst.isEndCap():
            # FIXME: regular endcaps
            assert any(
                master_name.startswith(p) for p in pad_name_prefixes
            ), master_name
            print("Found pad instance", inst_name, "of type", master_name)
            print("Found corner pad instance", inst_name, "of type", master_name)
            used_corner_pads.append((inst_name, master_name))
        else:
            assert not any(
                master_name.startswith(p) for p in pad_name_prefixes
            ), master_name
            other_instances.append(inst_name)

    # FIXME: if used_corner_pads aren't supposed to be instantiated
    assert len(used_corner_pads) == 4, used_corner_pads

    print()
    print(
        "The user design contains",
        len(used_pads),
        "pads, 4 corner pads, and",
        len(other_instances),
        "other instances",
    )
    print()
    assert len(used_pads) != 0, "^"

    ## Step 4: Generate a CFG or verify a user-provided config

    if config_file_name is not None:
        assert os.path.exists(config_file_name), config_file_name + " doesn't exist"
        with open(config_file_name, "r") as f:
            lines = f.readlines()
        user_config_pads = []
        for line in lines:
            if line.startswith("CORNER") or line.startswith("PAD"):
                tokens = line.split()
                assert len(tokens) == 5, tokens
                inst_name, master_name = tokens[1], tokens[3]
                if (
                    not pads[master_name] == "PAD_SPACER"
                    and not pads[master_name] == "PAD_AREAIO"
                ):
                    user_config_pads.append((inst_name, master_name))
            elif line.startswith("AREA"):
                tokens = line.split()
                assert len(tokens) == 4, tokens
                width = int(tokens[1])
                height = int(tokens[2])

        assert sorted(user_config_pads) == sorted(used_pads + used_corner_pads), (
            "Mismatch between the provided config and the provided netlist. Diff:",
            diff_lists(user_config_pads, used_pads + used_corner_pads),
        )

        print("User config verified")
        working_cfg = config_file_name
    else:
        # TODO: get minimum width/height so that --width and --height aren't required
        assert width is not None, "--width is required"
        assert height is not None, "--height is required"

        # auto generate a configuration

        # TODO: after calssification, center power pads on each side
        north, east, south, west = chunker(used_pads, 4)

        with open(working_cfg, "w") as f:
            f.write(
                generate_cfg(north, east, south, west, used_corner_pads, width, height)
            )

    if not init_only:
        invoke_padring(working_cfg, working_def, lefs)
    else:
        print(
            "Padframe config generated at",
            working_cfg,
            f"Modify it and re-run this program with the '-cfg {working_cfg}' option",
        )
        exit(0)

    print("Applying pad placements to the design DEF")

    padframe = OdbReader(lefs, working_def)

    assert padframe.name == "PADRING", padframe.name

    print("Padframe design name:", padframe.name)

    # Mark special nets
    if special_nets is not None:
        for net in top.block.getNets():
            net_name = net.getName()
            if net_name in special_nets:
                print("Marking", net_name, "as a special net")
                net.setSpecial()
                for iterm in net.getITerms():
                    iterm.setSpecial()

    # get minimum width/height (core-bounded)

    placed_cells_count = 0
    created_cells_count = 0
    for inst in padframe.block.getInsts():
        assert inst.isPad() or inst.isEndCap(), (
            inst.getName() + " is neither a pad nor corner pad"
        )

        inst_name = inst.getName()
        master = inst.getMaster()
        master_name = master.getName()
        x, y = inst.getLocation()
        orient = inst.getOrient()

        if (inst_name, master_name) in used_pads + used_corner_pads:
            original_inst = top.block.findInst(inst_name)
            assert original_inst is not None, "Failed to find " + inst_name
            assert original_inst.getPlacementStatus() == "NONE", (
                inst_name + " is already placed"
            )
            original_inst.setOrient(orient)
            original_inst.setLocation(x, y)
            original_inst.setPlacementStatus("FIRM")
            placed_cells_count += 1
        else:
            # must be a filler cell
            new_inst = odb.dbInst_create(
                top.block, top.db.findMaster(master_name), inst_name
            )
            assert new_inst is not None, "Failed to create " + inst_name
            new_inst.setOrient(orient)
            new_inst.setLocation(x, y)
            new_inst.setPlacementStatus("FIRM")
            created_cells_count += 1

    if output_def:
        odb.write_def(top.block, output_def)
    odb.write_db(top.db, output)
    print("Done.")


if __name__ == "__main__":
    padringer()



# Copyright 2020-2022 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import odb

import re
import math
import random

from reader import click_odb, click


def grid_to_tracks(origin, count, step):
    tracks = []
    pos = origin
    for _ in range(count):
        tracks.append(pos)
        pos += step
    assert len(tracks) > 0
    tracks.sort()

    return tracks


def equally_spaced_sequence(side, side_pin_placement, possible_locations):
    virtual_pin_count = 0
    actual_pin_count = len(side_pin_placement)
    total_pin_count = actual_pin_count + virtual_pin_count
    for i in range(len(side_pin_placement)):
        if isinstance(
            side_pin_placement[i], int
        ):  # This is an int value indicating virtual pins
            virtual_pin_count = virtual_pin_count + side_pin_placement[i]
            actual_pin_count = (
                actual_pin_count - 1
            )  # Decrement actual pin count, this value was only there to indicate virtual pin count
            total_pin_count = actual_pin_count + virtual_pin_count
    result = []
    tracks = len(possible_locations)

    if total_pin_count > tracks:
        print(
            f"There are more pins/virtual_pins: {total_pin_count}, than places to put them: {tracks}. Try making your floorplan area larger."
        )
        exit(1)
    elif total_pin_count == tracks:
        return possible_locations, side_pin_placement  # All positions.
    elif total_pin_count == 0:
        return result, side_pin_placement

    # From this point, pin_count always < tracks.
    tracks_per_pin = math.floor(tracks / total_pin_count)  # >=1
    # O| | | O| | | O| | |
    # tracks_per_pin = 3
    # notice the last two tracks are unused
    # thus:
    used_tracks = tracks_per_pin * (total_pin_count - 1) + 1
    unused_tracks = tracks - used_tracks

    # Place the pins at those tracks...
    current_track = unused_tracks // 2  # So that the tracks used are centered
    starting_track_index = current_track
    if virtual_pin_count == 0:  # No virtual pins
        for _ in range(0, total_pin_count):
            result.append(possible_locations[current_track])
            current_track += tracks_per_pin
    else:  # There are virtual pins
        for i in range(len(side_pin_placement)):
            if not isinstance(side_pin_placement[i], int):  # We have an actual pin
                result.append(possible_locations[current_track])
                current_track += tracks_per_pin
            else:  # Virtual Pins, so just leave their needed spaces
                current_track += tracks_per_pin * side_pin_placement[i]
        side_pin_placement = [
            pin for pin in side_pin_placement if not isinstance(pin, int)
        ]  # Remove the virtual pins from the side_pin_placement list

    print(f"Placement details for the {side} side")
    print("Virtual pin count: ", virtual_pin_count)
    print("Actual pin count: ", actual_pin_count)
    print("Total pin count: ", total_pin_count)
    print("Tracks count: ", len(possible_locations))
    print("Tracks per pin: ", tracks_per_pin)
    print("Used tracks count: ", used_tracks)
    print("Unused track count: ", unused_tracks)
    print("Starting track index: ", starting_track_index)

    VISUALIZE_PLACEMENT = True
    if VISUALIZE_PLACEMENT:
        print("Placement Map:")
        print("[", end="")
        used_track_indices = []
        for i, location in enumerate(possible_locations):
            if location in result:
                print(f"\033[91m{location}\033[0m, ", end="")
                used_track_indices.append(i)
            else:
                print(f"{location}, ", end="")
        print("]")
        print(f"Indices of used tracks: {used_track_indices}")
        print("---")

    return result, side_pin_placement


# HUMAN SORTING: https://stackoverflow.com/questions/5967500/how-to-correctly-sort-a-string-with-a-number-inside
def natural_keys(enum):
    def atof(text):
        try:
            retval = float(text)
        except ValueError:
            retval = text
        return retval

    text = enum[0]
    text = re.sub(r"(\[|\]|\.|\$)", "", text)
    """
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    (see toothy's implementation in the comments)
    float regex comes from https://stackoverflow.com/a/12643073/190597
    """
    return [atof(c) for c in re.split(r"[+-]?([0-9]+(?:[.][0-9]*)?|[.][0-9]+)", text)]


def bus_keys(enum):
    text = enum[0]
    m = re.match(r"^.*\[(\d+)\]$", text)
    if not m:
        return -1
    else:
        return int(m.group(1))


@click.command()
@click.option(
    "-u/-U",
    "--unmatched-error/--ignore-unmatched",
    default=False,
    help="Treat unmatched pins as error",
)
@click.option("-c", "--config", required=False, help="Optional configuration file.")
@click.option(
    "-r",
    "--reverse",
    default="",
    type=str,
    help="Reverse along comma,delimited,cardinals: e.g. N,E",
)
@click.option("-L", "--length", default=2, type=float, help="Pin length in microns.")
@click.option(
    "-V",
    "--ver-layer",
    required=True,
    help="Name of metal layer to place vertical pins on.",
)
@click.option(
    "-H",
    "--hor-layer",
    required=True,
    help="Name of metal layer to place horizontal pins on.",
)
@click.option(
    "--hor-extension",
    default=0,
    type=float,
    help="Extension for vertical pins in microns.",
)
@click.option(
    "--ver-extension",
    default=0,
    type=float,
    help="Extension for horizontal pins in microns.",
)
@click.option(
    "--ver-width-mult", default=2, type=float, help="Multiplier for vertical pins."
)
@click.option(
    "--hor-width-mult", default=2, type=float, help="Multiplier for horizontal pins."
)
@click.option(
    "--bus-sort/--no-bus-sort",
    default=False,
    help="Misnomer: pins are grouped by index instead of bus, i.e. a[0] goes with b[0] instead of a[1].",
)
@click_odb
def io_place(
    reader,
    config,
    ver_layer,
    hor_layer,
    ver_width_mult,
    hor_width_mult,
    length,
    hor_extension,
    ver_extension,
    reverse,
    bus_sort,
    unmatched_error,
):
    """
    Places the IOs in an input def with an optional config file that supports regexes.

    Config format:
    #N|#S|#E|#W
    pin1_regex (low co-ordinates to high co-ordinates; e.g., bottom to top and left to right)
    pin2_regex
    ...

    #S|#N|#E|#W
    """
    config_file_name = config
    bus_sort_flag = bus_sort
    unmatched_error_flag = unmatched_error

    h_layer_name = hor_layer
    v_layer_name = ver_layer

    h_width_mult = float(hor_width_mult)
    v_width_mult = float(ver_width_mult)

    micron_in_units = reader.dbunits

    LENGTH = int(micron_in_units * length)

    H_EXTENSION = int(micron_in_units * hor_extension)
    V_EXTENSION = int(micron_in_units * ver_extension)

    if H_EXTENSION < 0:
        H_EXTENSION = 0

    if V_EXTENSION < 0:
        V_EXTENSION = 0

    reverse_arr_raw = reverse.split(",")
    reverse_arr = []
    for element in reverse_arr_raw:
        if element.strip() != "":
            reverse_arr.append(f"#{element}")
    # read config

    pin_placement_cfg = {"#N": [], "#E": [], "#S": [], "#W": []}
    cur_side = None
    if config_file_name is not None and config_file_name != "":
        with open(config_file_name, "r") as config_file:
            for line in config_file:
                line = line.split()
                if len(line) == 0:
                    continue

                if len(line) > 1:
                    print("Only one entry allowed per line.")
                    exit(1)

                token = line[0]

                if cur_side is not None and token[0] != "#":
                    pin_placement_cfg[cur_side].append(token)
                elif token not in [
                    "#N",
                    "#E",
                    "#S",
                    "#W",
                    "#NR",
                    "#ER",
                    "#SR",
                    "#WR",
                    "#BUS_SORT",
                ]:
                    print(
                        "Valid directives are #N, #E, #S, or #W. Append R for reversing the default order.",
                        "Use #BUS_SORT to group 'bus bits' by index.",
                        "Please make sure you have set a valid side first before listing pins",
                    )
                    exit(1)
                elif token == "#BUS_SORT":
                    bus_sort_flag = True
                else:
                    if len(token) == 3:
                        token = token[0:2]
                        reverse_arr.append(token)
                    cur_side = token

    # build a list of pins
    H_LAYER = reader.tech.findLayer(h_layer_name)
    V_LAYER = reader.tech.findLayer(v_layer_name)

    H_WIDTH = int(h_width_mult * H_LAYER.getWidth())
    V_WIDTH = int(v_width_mult * V_LAYER.getWidth())

    print("Top-level design name:", reader.name)

    bterms = reader.block.getBTerms()
    bterms_enum = []
    for bterm in bterms:
        pin_name = bterm.getName()
        bterms_enum.append((pin_name, bterm))

    # sort them "humanly"
    bterms_enum.sort(key=natural_keys)
    if bus_sort_flag:
        bterms_enum.sort(key=bus_keys)
    bterms = [bterm[1] for bterm in bterms_enum]

    pin_placement = {"#N": [], "#E": [], "#S": [], "#W": []}
    bterm_regex_map = {}
    for side in pin_placement_cfg:
        for regex in pin_placement_cfg[side]:  # going through them in order
            if regex[0] == "$":  # Sign of Virtual Pins
                try:
                    virtual_pins_count = int(regex[1:])
                    pin_placement[side].append(virtual_pins_count)
                except ValueError:
                    print("You provided invalid values for virtual pins")
                    exit(1)
            else:
                regex += "$"  # anchor
                for bterm in bterms:
                    # if a pin name matches multiple regexes, their order will be
                    # arbitrary. More refinement requires more strict regexes (or just
                    # the exact pin name).
                    pin_name = bterm.getName()
                    if re.match(regex, pin_name) is not None:
                        if bterm in bterm_regex_map:
                            print(
                                f"Error: Multiple regexes matched {pin_name}. Those are {bterm_regex_map[bterm]} and {regex}"
                            )
                            exit(1)
                        bterm_regex_map[bterm] = regex
                        pin_placement[side].append(bterm)  # to maintain the order

    unmatched_bterms = [bterm for bterm in bterms if bterm not in bterm_regex_map]

    if len(unmatched_bterms) > 0:
        print("Warning: Some pins weren't matched by the config file")
        print("Those are:", [bterm.getName() for bterm in unmatched_bterms])
        if unmatched_error_flag:
            print("Treating unmatched pins as errors. Exiting..")
            exit(1)
        else:
            print("Assigning random sides to the above pins")
            for bterm in unmatched_bterms:
                random_side = random.choice(list(pin_placement.keys()))
                pin_placement[random_side].append(bterm)

    # generate slots

    DIE_AREA = reader.block.getDieArea()
    BLOCK_LL_X = DIE_AREA.xMin()
    BLOCK_LL_Y = DIE_AREA.yMin()
    BLOCK_UR_X = DIE_AREA.xMax()
    BLOCK_UR_Y = DIE_AREA.yMax()

    print("Block boundaries:", BLOCK_LL_X, BLOCK_LL_Y, BLOCK_UR_X, BLOCK_UR_Y)

    origin, count, step = reader.block.findTrackGrid(H_LAYER).getGridPatternY(0)
    print(f"Horizontal Tracks Origin: {origin}, Count: {count}, Step: {step}")
    h_tracks = grid_to_tracks(origin, count, step)

    origin, count, step = reader.block.findTrackGrid(V_LAYER).getGridPatternX(0)
    print(f"Vertical Tracks Origin: {origin}, Count: {count}, Step: {step}")
    v_tracks = grid_to_tracks(origin, count, step)

    for rev in reverse_arr:
        pin_placement[rev].reverse()

    # create the pins
    for side in pin_placement:
        if side in ["#N", "#S"]:
            slots, pin_placement[side] = equally_spaced_sequence(
                side, pin_placement[side], v_tracks
            )
        else:
            slots, pin_placement[side] = equally_spaced_sequence(
                side, pin_placement[side], h_tracks
            )

        assert len(slots) == len(pin_placement[side])

        for i in range(len(pin_placement[side])):
            bterm = pin_placement[side][i]
            slot = slots[i]
            # print(f"Pin name: {bterm.getName()}, placed at slot: {slot}")

            pin_name = bterm.getName()
            pins = bterm.getBPins()
            if len(pins) > 0:
                # print(f"Warning: {pin_name} already has shapes. Modifying them")
                assert (
                    len(pins) == 1
                ), f"{pin_name} has multiple pins already placed. This is a fatal error."
                pin_bpin = pins[0]
            else:
                pin_bpin = odb.dbBPin_create(bterm)

            pin_bpin.setPlacementStatus("PLACED")

            if side in ["#N", "#S"]:
                rect = odb.Rect(0, 0, V_WIDTH, LENGTH + V_EXTENSION)
                if side == "#N":
                    y = BLOCK_UR_Y - LENGTH
                else:
                    y = BLOCK_LL_Y - V_EXTENSION
                rect.moveTo(slot - V_WIDTH // 2, y)
                odb.dbBox_create(pin_bpin, V_LAYER, *rect.ll(), *rect.ur())
            else:
                rect = odb.Rect(0, 0, LENGTH + H_EXTENSION, H_WIDTH)
                if side == "#E":
                    x = BLOCK_UR_X - LENGTH
                else:
                    x = BLOCK_LL_X - H_EXTENSION
                rect.moveTo(x, slot - H_WIDTH // 2)
                odb.dbBox_create(pin_bpin, H_LAYER, *rect.ll(), *rect.ur())


if __name__ == "__main__":
    io_place()



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import json
from glob import glob
from decimal import Decimal
from typing import List, Sequence, Tuple, Optional, Callable, Union, Dict

from .resolve import resolve, Keys
from .pdk import (
    all_variables as pdk_variables,
    removed_variables as pdk_removed_variables,
    migrate_old_config,
)
from .flow import removed_variables, all_variables as flow_common_variables
from .tcleval import env_from_tcl
from .config import Config, Meta
from .variable import Variable
from ..logging import info, warn


class InvalidConfig(ValueError):
    """
    An error raised when a configuration under resolution is invalid.

    :param config: A human-readable name for the particular configuration file
        causing this exception, i.e. whether it's a PDK configuration file or a
        user configuration file.
    :param warnings: A list of warnings generated during the loading of this
        configuration file.
    :param errors: A list of errors generated during the loading of this
        configuration file.
    :param args: Further arguments to be passed onto the constructor of
        :class:`ValueError`.
    :param kwargs: Further keyword arguments to be passed onto the constructor of
        :class:`ValueError`.
    """

    def __init__(
        self,
        config: str,
        warnings: List[str],
        errors: List[str],
        *args,
        **kwargs,
    ) -> None:
        self.config = config
        self.warnings = warnings
        self.errors = errors
        super().__init__(*args, **kwargs)


class DecimalDecoder(json.JSONDecoder):
    def default(self, o):
        if isinstance(o, float) or isinstance(o, int):
            return Decimal(o)
        return super(DecimalDecoder, self).default(o)


class ConfigBuilder(object):
    """
    Various constructors for OpenLane configuration objects.
    """

    @classmethod
    def interactive(
        Self,
        DESIGN_NAME: str,
        PDK: str,
        STD_CELL_LIBRARY: Optional[str] = None,
        PDK_ROOT: Optional[str] = None,
        **kwargs,
    ) -> Config:
        """
        This constructs a partial configuration object that may be incrementally
        adjusted per-step, and activates OpenLane's **interactive mode**.

        The interactive mode is overall less rigid than the pure mode, adding various
        references to global objects to make the REPL or Notebook experience more
        pleasant, however, it is not as resilient as the pure mode and should not
        be used in production code.

        :param DESIGN_NAME: The name of the design to be used.
        :param PDK: The name of the PDK.
        :param STD_CELL_LIBRARY: The name of the standard cell library.

            If not specified, the PDK's default SCL will be used.
        :param PDK_ROOT: Required if Volare is not installed.

            If Volare is installed, this value can be used to optionally override
            Volare's default.

        :param kwargs: Any overrides to PDK values and/or common flow default variables
            can be passed as keyword arguments to this function.

            Useful examples are CLOCK_PORT, CLOCK_PERIOD, et cetera, which while
            not bound to a specific :class:`Step`, affects most Steps' behavior.
        """
        config_in, _, _ = Self._get_pdk_config(
            PDK,
            STD_CELL_LIBRARY,
            PDK_ROOT,
        )

        config_in["PDK_ROOT"] = PDK_ROOT
        config_in["DESIGN_NAME"] = DESIGN_NAME

        config_in.update(kwargs)

        config_in, design_warnings, design_errors = Variable.process_config(
            config_in,
            pdk_variables + list(flow_common_variables),
            removed_variables,
        )

        if len(design_errors) != 0:
            raise InvalidConfig("default configuration", design_warnings, design_errors)

        if len(design_warnings) > 0:
            info(
                "Loading the default configuration has generated the following warnings:"
            )
        for warning in design_warnings:
            warn(warning)

        config_in.interactive = True

        Config.current_interactive = config_in

        return config_in

    @classmethod
    def load(
        Self,
        config_in: Union[str, os.PathLike, Dict],
        flow_config_vars: Sequence[Variable],
        config_override_strings: Optional[Sequence[str]] = None,
        pdk: Optional[str] = None,
        pdk_root: Optional[str] = None,
        scl: Optional[str] = None,
        design_dir: Optional[str] = None,
    ) -> Tuple["Config", str]:
        """
        :param config_in: Either a file path to a JSON file or a Python
            dictionary representing an unprocessed OpenLane configuration
            object.

            Tcl files are also supported, but are deprecated and will be removed
            in the future.

        :param config_override_strings: A list of "overrides" in the form of
            NAME=VALUE strings. These are primarily for running OpenLane from
            the commandline and strictly speaking should not be used in the API.

        :param design_dir: The design directory for said configuration.
            Supported and required *if and only if* config_in is a dictionary.

        :param pdk: A process design kit to use. Required unless specified via the
            "PDK" key in a configuration object.

        :param pdk_root: Required if Volare is not installed.

            If Volare is installed, this value can be used to optionally override
            Volare's default.

        :param scl: A standard cell library to use. If not specified, the PDK's
            default standard cell library will be used instead.

        :returns: A tuple containing a ConfigBuilder and the design directory.
        """

        loader: Callable = Self._loads
        raw: Union[str, dict] = ""
        if not isinstance(config_in, dict):
            if design_dir is not None:
                raise TypeError(
                    "The argument design_dir is not supported when config_in is not a dictionary."
                )

            design_dir = str(os.path.dirname(config_in))
            config_in = str(config_in)
            if config_in.endswith(".json"):
                raw = open(config_in, encoding="utf8").read()
            elif config_in.endswith(".tcl"):
                raw = open(config_in, encoding="utf8").read()
                loader = Self._loads_tcl
                if config_override_strings is None:
                    raise ValueError(
                        "CLI override strings are not supported with .Tcl configuration files."
                    )
            else:
                if os.path.isdir(config_in):
                    raise ValueError(
                        "Passing design folders as arguments is unsupported in OpenLane 2.0+: please pass the JSON configuration file directly."
                    )
                _, ext = os.path.splitext(config_in)
                raise ValueError(
                    f"Unsupported configuration file extension '{ext}' for '{config_in}'."
                )
        else:
            if design_dir is None:
                raise TypeError(
                    "The argument design_dir is required when using ConfigBuilder with a dictionary."
                )
            raw = config_in
            loader = Self._load_dict

        loaded = loader(
            raw,
            design_dir,
            flow_config_vars=flow_config_vars,
            pdk=pdk,
            pdk_root=pdk_root,
            scl=scl,
            config_override_strings=(config_override_strings or []),
        )

        return (loaded, design_dir)

    @classmethod
    def _loads(
        Self,
        json_str: str,
        *args,
        **kwargs,
    ):
        raw = json.loads(json_str, cls=DecimalDecoder)
        kwargs["resolve_json"] = True
        return Self._load_dict(
            raw,
            *args,
            **kwargs,
        )

    @classmethod
    def _load_dict(
        Self,
        raw: dict,
        design_dir: str,
        flow_config_vars: Sequence[Variable],
        config_override_strings: Sequence[str],
        pdk: Optional[str] = None,
        pdk_root: Optional[str] = None,
        scl: Optional[str] = None,
        full_pdk_warnings: bool = False,
        resolve_json: bool = False,
    ) -> "Config":
        meta_raw: Optional[dict] = None
        if raw.get("meta") is not None:
            meta_raw = raw["meta"]
            del raw["meta"]

        for string in config_override_strings:
            key, value = string.split("=", 1)
            raw[key] = value

        process_info = resolve(
            raw,
            only_extract_process_info=True,
            design_dir=design_dir,
        )

        pdk = process_info.get(Keys.pdk) or pdk
        if pdk is None:
            raise ValueError(
                "The pdk argument is required as the configuration object lacks a 'PDK' key."
            )

        config_in, pdkpath, scl = Self._get_pdk_config(
            pdk=pdk,
            scl=scl,
            pdk_root=pdk_root,
            full_pdk_warnings=full_pdk_warnings,
        )

        resolve_maybe = resolve if resolve_json else lambda x, *args, **kwargs: x

        resolved = resolve_maybe(
            raw,
            pdk=pdk,
            pdkpath=pdkpath,
            scl=config_in["STD_CELL_LIBRARY"],
            design_dir=design_dir,
        )

        config_in.update(**resolved)

        config_in, design_warnings, design_errors = Variable.process_config(
            config_in,
            pdk_variables + list(flow_config_vars),
            removed_variables,
        )

        if meta_raw is not None:
            try:
                config_in.meta = Meta(**meta_raw)
            except TypeError as e:
                design_errors.append(f"'meta' object is invalid: {e}")

        if len(design_errors) != 0:
            raise InvalidConfig(
                "design configuration file", design_warnings, design_errors
            )

        if len(design_warnings) > 0:
            info(
                "Loading the design configuration file has generated the following warnings:"
            )
        for warning in design_warnings:
            warn(warning)

        config_in["PDK_ROOT"] = pdk_root
        return config_in

    @classmethod
    def _loads_tcl(
        Self,
        config: str,
        design_dir: str,
        flow_config_vars: Sequence[Variable],
        config_override_strings: Sequence[str],  # Unused, kept for API consistency
        pdk: Optional[str] = None,
        pdk_root: Optional[str] = None,
        scl: Optional[str] = None,
        full_pdk_warnings: bool = False,
    ) -> "Config":
        warn(
            "Support for .tcl configuration files is deprecated. Please migrate to a .json file at your earliest convenience."
        )

        try:
            import volare

            pdk_root = volare.get_volare_home(pdk_root)
        except ImportError:
            if pdk_root is None:
                raise ValueError(
                    "The pdk_root argument is required as Volare is not installed."
                )

        config_in = Config(
            {
                "PDK_ROOT": pdk_root,
                Keys.pdk: pdk,
            }
        )

        tcl_vars_in = config_in.copy()
        tcl_vars_in[Keys.scl] = ""
        tcl_vars_in[Keys.design_dir] = os.path.abspath(design_dir)
        tcl_config = env_from_tcl(tcl_vars_in, config)

        process_info = resolve(
            tcl_config.data,
            only_extract_process_info=True,
            design_dir=design_dir,
        )

        pdk = process_info.get(Keys.pdk) or pdk

        if pdk is None:
            raise ValueError(
                "The pdk argument is required as the configuration object lacks a 'PDK' key."
            )

        config_in, _, scl = Self._get_pdk_config(
            pdk=pdk,
            scl=scl,
            pdk_root=pdk_root,
            full_pdk_warnings=full_pdk_warnings,
        )

        tcl_vars_in[Keys.pdk] = pdk
        tcl_vars_in[Keys.scl] = scl
        tcl_vars_in[Keys.design_dir] = os.path.abspath(design_dir)

        design_config = env_from_tcl(tcl_vars_in, config)

        config_in.update(**design_config)

        config_in, design_warnings, design_errors = Variable.process_config(
            config_in,
            pdk_variables + list(flow_config_vars),
            removed_variables,
        )

        if len(design_errors) != 0:
            raise InvalidConfig(
                "design configuration file", design_warnings, design_errors
            )

        if len(design_warnings) > 0:
            info(
                "Loading the design configuration file has generated the following warnings:"
            )
        for warning in design_warnings:
            warn(warning)

        config_in["PDK_ROOT"] = os.path.abspath(pdk_root)

        return config_in

    @classmethod
    def _get_pdk_config(
        Self,
        pdk: str,
        scl: Optional[str] = None,
        pdk_root: Optional[str] = None,
        full_pdk_warnings: Optional[bool] = False,
    ) -> Tuple[Config, str, str]:
        """
        :returns: A tuple of the PDK configuration, the PDK path and the SCL.
        """
        try:
            import volare

            pdk_root = volare.get_volare_home(pdk_root)
        except ImportError:
            if pdk_root is None:
                raise ValueError(
                    "The pdk_root argument is required as Volare is not installed."
                )

        config_in = Config(
            {
                "PDK_ROOT": pdk_root,
                Keys.pdk: pdk,
            }
        )
        if scl is not None:
            config_in[Keys.scl] = scl

        pdkpath = os.path.join(pdk_root, pdk)
        if not os.path.exists(pdkpath):
            matches = glob(f"{pdkpath}*")
            errors = [f"The PDK {pdk} was not found."]
            warnings = []
            for match in matches:
                basename = os.path.basename(match)
                warnings.append(f"A similarly-named PDK was found: {basename}")
            raise InvalidConfig("PDK configuration", warnings, errors)

        pdk_config_path = os.path.join(pdkpath, "libs.tech", "openlane", "config.tcl")

        config_in = env_from_tcl(
            config_in,
            open(pdk_config_path, encoding="utf8").read(),
        )

        scl = config_in["STD_CELL_LIBRARY"]
        assert (
            scl is not None
        ), "Fatal error: STD_CELL_LIBRARY default value not set by PDK."

        scl_config_path = os.path.join(
            pdkpath, "libs.tech", "openlane", scl, "config.tcl"
        )
        config_in = migrate_old_config(
            env_from_tcl(
                config_in,
                open(scl_config_path, encoding="utf8").read(),
            )
        )

        config_in, pdk_warnings, pdk_errors = Variable.process_config(
            config_in,
            pdk_variables,
            pdk_removed_variables,
        )

        if len(pdk_errors) != 0:
            raise InvalidConfig("PDK configuration files", pdk_warnings, pdk_errors)

        if len(pdk_warnings) > 0:
            if full_pdk_warnings:
                info(
                    "Loading the PDK configuration files has generated the following warnings:"
                )
                for warning in pdk_warnings:
                    warn(warning)

        return (config_in, pdkpath, scl)
